# 1 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2018  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief General methods for testing DBCSR tensors.
!> \author Patrick Seewald
! **************************************************************************************************
MODULE dbcsr_tensor_test

# 1 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor.fypp" 1
# 5 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor.fypp"
# 53 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor.fypp"
# 13 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F" 2
# 14 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 15 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

   USE dbcsr_api,                       ONLY: dbcsr_type_real_8, dbcsr_type_real_4, dbcsr_type_complex_4, dbcsr_type_complex_8, &
                                              dbcsr_get_info
   USE dbcsr_tensor,                    ONLY: &
        dbcsr_t_copy, dbcsr_t_get_block, dbcsr_t_iterator_type, dbcsr_t_iterator_blocks_left, &
        dbcsr_t_iterator_next_block, dbcsr_t_iterator_start, dbcsr_t_iterator_stop, dbcsr_t_ndims, &
        dbcsr_t_reserve_blocks, dbcsr_t_get_stored_coordinates, dbcsr_t_put_block, &
        dbcsr_t_contract, dbcsr_t_inverse_order, dbcsr_t_dims
   USE dbcsr_tensor_block,              ONLY: block_nd
   USE dbcsr_tensor_types,              ONLY: dbcsr_t_create,&
                                              dbcsr_t_destroy,&
                                              dbcsr_t_type,&
                                              dbcsr_t_distribution_type,&
                                              dbcsr_t_distribution_destroy,&
                                              dims_tensor,&
                                              dbcsr_t_distribution_new,&
                                              dbcsr_t_nd_mp_comm,&
                                              dbcsr_t_get_data_type
   USE kinds,                           ONLY: real_4, real_8,&
                                              default_string_length,&
                                              int_8
   USE message_passing,                 ONLY: mp_environ, &
                                              mp_bcast, &
                                              mp_comm_free, &
                                              mp_comm_free, &
                                              mp_sum
   USE allocate_wrap,                   ONLY: allocate_any
   USE dbcsr_tensor_index,              ONLY: combine_index, &
                                              get_mapping_info
#include "../base/base_uses.f90"

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_tensor_test'

   PUBLIC :: &
      dbcsr_t_setup_test_tensor,&
      dbcsr_t_random_dist,&
      dbcsr_t_contract_test,&
      dbcsr_t_test_formats

   INTERFACE repl_dense_array_to_dist_sparse_tensor
# 58 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 59 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE repl_dense_2d_array_to_dist_sparse_tensor_r_dp
# 59 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE repl_dense_3d_array_to_dist_sparse_tensor_r_dp
# 59 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE repl_dense_4d_array_to_dist_sparse_tensor_r_dp
# 61 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 58 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 59 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE repl_dense_2d_array_to_dist_sparse_tensor_r_sp
# 59 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE repl_dense_3d_array_to_dist_sparse_tensor_r_sp
# 59 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE repl_dense_4d_array_to_dist_sparse_tensor_r_sp
# 61 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 58 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 59 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE repl_dense_2d_array_to_dist_sparse_tensor_c_dp
# 59 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE repl_dense_3d_array_to_dist_sparse_tensor_c_dp
# 59 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE repl_dense_4d_array_to_dist_sparse_tensor_c_dp
# 61 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 58 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 59 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE repl_dense_2d_array_to_dist_sparse_tensor_c_sp
# 59 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE repl_dense_3d_array_to_dist_sparse_tensor_c_sp
# 59 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE repl_dense_4d_array_to_dist_sparse_tensor_c_sp
# 61 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 62 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
   END INTERFACE

   INTERFACE write_array
# 66 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE write_2d_array
# 66 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE write_3d_array
# 66 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE write_4d_array
# 68 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
   END INTERFACE

   INTERFACE dist_sparse_tensor_to_repl_dense_array
# 72 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 73 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE dist_sparse_tensor_to_repl_dense_2d_array_r_dp
# 73 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE dist_sparse_tensor_to_repl_dense_3d_array_r_dp
# 73 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE dist_sparse_tensor_to_repl_dense_4d_array_r_dp
# 75 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 72 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 73 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE dist_sparse_tensor_to_repl_dense_2d_array_r_sp
# 73 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE dist_sparse_tensor_to_repl_dense_3d_array_r_sp
# 73 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE dist_sparse_tensor_to_repl_dense_4d_array_r_sp
# 75 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 72 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 73 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE dist_sparse_tensor_to_repl_dense_2d_array_c_dp
# 73 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE dist_sparse_tensor_to_repl_dense_3d_array_c_dp
# 73 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE dist_sparse_tensor_to_repl_dense_4d_array_c_dp
# 75 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 72 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 73 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE dist_sparse_tensor_to_repl_dense_2d_array_c_sp
# 73 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE dist_sparse_tensor_to_repl_dense_3d_array_c_sp
# 73 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      MODULE PROCEDURE dist_sparse_tensor_to_repl_dense_4d_array_c_sp
# 75 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 76 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
   END INTERFACE
CONTAINS

! **************************************************************************************************
!> \brief check if two (arbitrarily mapped and distributed) tensors are equal.
!> \param tensor1 ...
!> \param tensor2 ...
!> \return ...
! **************************************************************************************************
   FUNCTION dbcsr_t_equal(tensor1, tensor2)
      TYPE(dbcsr_t_type), INTENT(INOUT)          :: tensor1, tensor2
      LOGICAL                                    :: dbcsr_t_equal

      INTEGER                                    :: blk
      TYPE(dbcsr_t_type)                         :: tensor2_tmp
      TYPE(dbcsr_t_iterator_type)                :: iter
      TYPE(block_nd)                             :: blk_data1, blk_data2
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor1)) :: blk_size, ind_nd

      ! create a copy of tensor2 that has exact same data format as tensor1
      CALL dbcsr_t_create(tensor1, tensor2_tmp)
      CALL dbcsr_t_reserve_blocks(tensor1, tensor2_tmp)
      CALL dbcsr_t_copy(tensor2, tensor2_tmp)

      dbcsr_t_equal = .TRUE.

      CALL dbcsr_t_iterator_start(iter, tensor1)

      DO WHILE (dbcsr_t_iterator_blocks_left(iter))
         CALL dbcsr_t_iterator_next_block(iter, ind_nd, blk, blk_size=blk_size)
         CALL dbcsr_t_get_block(tensor1, ind_nd, blk_data1)
         CALL dbcsr_t_get_block(tensor2_tmp, ind_nd, blk_data2)

         IF (.NOT. blocks_equal(blk_data1, blk_data2)) THEN
            dbcsr_t_equal = .FALSE.
         ENDIF
      ENDDO

      CALL dbcsr_t_iterator_stop(iter)
      CALL dbcsr_t_destroy(tensor2_tmp)
   END FUNCTION

! **************************************************************************************************
!> \brief check if two blocks are equal
!> \param block1 ...
!> \param block2 ...
!> \return ...
! **************************************************************************************************
   PURE FUNCTION blocks_equal(block1, block2)
      TYPE(block_nd), INTENT(IN) :: block1, block2
      LOGICAL                    :: blocks_equal

      SELECT CASE (block1%data_type)
# 130 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE (dbcsr_type_real_8)
         blocks_equal = MAXVAL(ABS(block1%r_dp%blk-block2%r_dp%blk)) .LT. 1.0E-12_real_8
# 130 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE (dbcsr_type_real_4)
         blocks_equal = MAXVAL(ABS(block1%r_sp%blk-block2%r_sp%blk)) .LT. 1.0E-12_real_4
# 130 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE (dbcsr_type_complex_8)
         blocks_equal = MAXVAL(ABS(block1%c_dp%blk-block2%c_dp%blk)) .LT. 1.0E-12_real_8
# 130 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE (dbcsr_type_complex_4)
         blocks_equal = MAXVAL(ABS(block1%c_sp%blk-block2%c_sp%blk)) .LT. 1.0E-12_real_4
# 133 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      END SELECT

   END FUNCTION

! **************************************************************************************************
!> \brief Compute factorial
!> \param n ...
!> \return ...
! **************************************************************************************************
   PURE FUNCTION factorial(n)
      INTEGER, INTENT(IN) :: n
      INTEGER             :: k
      INTEGER             :: factorial
      factorial = PRODUCT((/(k, k=1, n)/))
   END FUNCTION

! **************************************************************************************************
!> \brief Compute all permutations p of (1, 2, ..., n)
!> \param n ...
!> \param p ...
! **************************************************************************************************
   SUBROUTINE permute(n, p)
      INTEGER, INTENT(IN)                              :: n
      INTEGER                                          :: i, c
      INTEGER, DIMENSION(n)                            :: pp
      INTEGER, DIMENSION(n, factorial(n)), INTENT(OUT) :: p

      pp = [(i, i=1, n)]
      c = 1
      CALL perm(1)
   CONTAINS
      RECURSIVE SUBROUTINE perm(i)
         INTEGER, INTENT(IN) :: i
         INTEGER :: j, t
         IF (i == n) THEN
            p(:, c) = pp(:)
            c = c+1
         ELSE
            DO j = i, n
               t = pp(i)
               pp(i) = pp(j)
               pp(j) = t
               call perm(i+1)
               t = pp(i)
               pp(i) = pp(j)
               pp(j) = t
            END DO
         END IF
      END SUBROUTINE
   END SUBROUTINE

! **************************************************************************************************
!> \brief Test equivalence of all tensor formats, using a random distribution.
!> \param ndims tensor rank
!> \param mp_comm_2d communicator
!> \param unit_nr output unit, needs to be a valid unit number on all mpi ranks
!> \param verbose if .TRUE., print all tensor blocks
!> \param blk_size_1 block sizes along dimension 1
!> \param blk_ind_1 index along dimension 1 of non-zero blocks
! **************************************************************************************************
   SUBROUTINE dbcsr_t_test_formats(ndims, mp_comm_2d, unit_nr, verbose, &
                                  blk_size_1, blk_size_2, blk_size_3, blk_size_4, &
                                  blk_ind_1, blk_ind_2, blk_ind_3, blk_ind_4)
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL :: blk_size_1, blk_size_2, blk_size_3, blk_size_4
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL :: blk_ind_1, blk_ind_2, blk_ind_3, blk_ind_4
      INTEGER, INTENT(IN)                         :: ndims
      INTEGER, INTENT(IN)                         :: unit_nr
      LOGICAL, INTENT(IN)                         :: verbose
      INTEGER, INTENT(IN)                         :: mp_comm_2d
      INTEGER                                     :: mp_comm_ref, mp_comm_test
      TYPE(dbcsr_t_distribution_type)             :: dist1, dist2
      TYPE(dbcsr_t_type)                          :: tensor1, tensor2
      INTEGER                                     :: isep, iblk
      INTEGER, DIMENSION(:), ALLOCATABLE          :: dist1_1, dist1_2, dist1_3, dist1_4, &
                                                     dist2_1, dist2_2, dist2_3, dist2_4
      INTEGER                                     :: nblks, imap
      INTEGER, DIMENSION(ndims)                   :: pdims_ref, pdims_test, myploc
      LOGICAL                                     :: eql
      INTEGER                                     :: iperm, idist, icount
      INTEGER, DIMENSION(:), ALLOCATABLE          :: map1, map2, map1_ref, map2_ref
      INTEGER, DIMENSION(ndims, factorial(ndims)) :: perm
      LOGICAL, DIMENSION(ndims)                   :: periods
      INTEGER                                     :: io_unit
      INTEGER                                     :: mynode, numnodes
      CHARACTER(LEN=default_string_length)        :: tensor_name

      CALL mp_environ(numnodes, mynode, mp_comm_2d)

      io_unit = 0
      IF (mynode .EQ. 0) io_unit = unit_nr

      CALL permute(ndims, perm)
      CALL allocate_any(map1_ref, source=perm(1:ndims/2, 1))
      CALL allocate_any(map2_ref, source=perm(ndims/2+1:ndims, 1))

      mp_comm_ref = dbcsr_t_nd_mp_comm(mp_comm_2d, map1_ref, map2_ref)
      CALL mp_environ(numnodes, mynode, mp_comm_ref)
      CALL mp_environ(mp_comm_ref, ndims, pdims_ref, myploc, periods)

      IF(io_unit > 0) THEN
         WRITE(io_unit, *)
         WRITE(io_unit, '(A)') repeat("-",80)
         WRITE(io_unit, '(A,1X,I1)') "Testing matrix representations of tensor rank", ndims
         WRITE(io_unit, '(A)') repeat("-",80)
         WRITE(io_unit,'(A)') "Block sizes:"

# 240 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF (ndims>=1) THEN
            WRITE(io_unit,'(T4,A,1X,I1,A,1X)', advance='no') 'Dim', 1,':'
            DO iblk=1, SIZE(blk_size_1)
               WRITE(io_unit,'(I2,1X)', advance='no') blk_size_1(iblk)
            ENDDO
            WRITE(io_unit,*)
         ENDIF
# 240 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF (ndims>=2) THEN
            WRITE(io_unit,'(T4,A,1X,I1,A,1X)', advance='no') 'Dim', 2,':'
            DO iblk=1, SIZE(blk_size_2)
               WRITE(io_unit,'(I2,1X)', advance='no') blk_size_2(iblk)
            ENDDO
            WRITE(io_unit,*)
         ENDIF
# 240 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF (ndims>=3) THEN
            WRITE(io_unit,'(T4,A,1X,I1,A,1X)', advance='no') 'Dim', 3,':'
            DO iblk=1, SIZE(blk_size_3)
               WRITE(io_unit,'(I2,1X)', advance='no') blk_size_3(iblk)
            ENDDO
            WRITE(io_unit,*)
         ENDIF
# 240 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF (ndims>=4) THEN
            WRITE(io_unit,'(T4,A,1X,I1,A,1X)', advance='no') 'Dim', 4,':'
            DO iblk=1, SIZE(blk_size_4)
               WRITE(io_unit,'(I2,1X)', advance='no') blk_size_4(iblk)
            ENDDO
            WRITE(io_unit,*)
         ENDIF
# 248 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

         WRITE(io_unit,'(A)') "Non-zero blocks:"
         DO iblk=1,SIZE(blk_ind_1)
# 252 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF (ndims==2) THEN
               WRITE(io_unit,'(T4,A, I3, A, 2I3, 1X, A)') &
                  'Block',iblk, ': (', blk_ind_1(iblk), blk_ind_2(iblk), ')'
            ENDIF
# 252 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF (ndims==3) THEN
               WRITE(io_unit,'(T4,A, I3, A, 3I3, 1X, A)') &
                  'Block',iblk, ': (', blk_ind_1(iblk), blk_ind_2(iblk), blk_ind_3(iblk), ')'
            ENDIF
# 252 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF (ndims==4) THEN
               WRITE(io_unit,'(T4,A, I3, A, 4I3, 1X, A)') &
                  'Block',iblk, ': (', blk_ind_1(iblk), blk_ind_2(iblk), blk_ind_3(iblk), blk_ind_4(iblk), ')'
            ENDIF
# 257 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         ENDDO

         WRITE(io_unit, *)
         WRITE(io_unit,'(A,1X)', advance='no') "Reference map:"
         WRITE (io_unit, '(A1,1X)', advance='no') "("
         DO imap = 1, SIZE(map1_ref)
            WRITE (io_unit, '(I1,1X)', advance='no') map1_ref(imap)
         ENDDO
         WRITE (io_unit, '(A1,1X)', advance='no') "|"
         DO imap = 1, SIZE(map2_ref)
            WRITE (io_unit, '(I1,1X)', advance='no') map2_ref(imap)
         ENDDO
         WRITE (io_unit, '(A1)') ")"

      ENDIF

      icount = 0
      DO iperm = 1, factorial(ndims)
         DO isep = 1, ndims-1
            icount = icount+1

            CALL allocate_any(map1, source=perm(1:isep,iperm))
            CALL allocate_any(map2, source=perm(isep+1:ndims,iperm))

            mp_comm_test = dbcsr_t_nd_mp_comm(mp_comm_2d, map1, map2)
            CALL mp_environ(numnodes, mynode, mp_comm_test)
            CALL mp_environ(mp_comm_test, ndims, pdims_test, myploc, periods)

# 286 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF (1 <= ndims) THEN
               nblks = SIZE(blk_size_1)
               CALL dbcsr_t_random_dist(dist1_1, nblks, pdims_test(1), mp_comm_test)
               CALL dbcsr_t_random_dist(dist2_1, nblks, pdims_ref(1), mp_comm_ref)
            ENDIF
# 286 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF (2 <= ndims) THEN
               nblks = SIZE(blk_size_2)
               CALL dbcsr_t_random_dist(dist1_2, nblks, pdims_test(2), mp_comm_test)
               CALL dbcsr_t_random_dist(dist2_2, nblks, pdims_ref(2), mp_comm_ref)
            ENDIF
# 286 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF (3 <= ndims) THEN
               nblks = SIZE(blk_size_3)
               CALL dbcsr_t_random_dist(dist1_3, nblks, pdims_test(3), mp_comm_test)
               CALL dbcsr_t_random_dist(dist2_3, nblks, pdims_ref(3), mp_comm_ref)
            ENDIF
# 286 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF (4 <= ndims) THEN
               nblks = SIZE(blk_size_4)
               CALL dbcsr_t_random_dist(dist1_4, nblks, pdims_test(4), mp_comm_test)
               CALL dbcsr_t_random_dist(dist2_4, nblks, pdims_ref(4), mp_comm_ref)
            ENDIF
# 292 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

            IF(io_unit > 0) THEN
               WRITE(io_unit, *)
               WRITE(tensor_name, '(A,1X,I3,1X)') "Test", icount
               WRITE(io_unit, '(A,A,1X)', advance='no') TRIM(tensor_name),':'
               WRITE (io_unit, '(A1,1X)', advance='no') "("
               DO imap = 1, SIZE(map1)
                  WRITE (io_unit, '(I1,1X)', advance='no') map1(imap)
               ENDDO
               WRITE (io_unit, '(A1,1X)', advance='no') "|"
               DO imap = 1, SIZE(map2)
                  WRITE (io_unit, '(I1,1X)', advance='no') map2(imap)
               ENDDO
               WRITE (io_unit, '(A1)') ")"

               WRITE(io_unit,'(T4,A)') "Reference distribution:"
# 309 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
               IF (1 <= ndims) THEN
                  WRITE(io_unit, '(T7,A,1X)', advance='no') "Dist vec 1:"
                  DO idist = 1, SIZE(dist2_1)
                     WRITE(io_unit, '(I2,1X)', advance='no') dist2_1(idist)
                  ENDDO
                  WRITE(io_unit, *)
               ENDIF
# 309 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
               IF (2 <= ndims) THEN
                  WRITE(io_unit, '(T7,A,1X)', advance='no') "Dist vec 2:"
                  DO idist = 1, SIZE(dist2_2)
                     WRITE(io_unit, '(I2,1X)', advance='no') dist2_2(idist)
                  ENDDO
                  WRITE(io_unit, *)
               ENDIF
# 309 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
               IF (3 <= ndims) THEN
                  WRITE(io_unit, '(T7,A,1X)', advance='no') "Dist vec 3:"
                  DO idist = 1, SIZE(dist2_3)
                     WRITE(io_unit, '(I2,1X)', advance='no') dist2_3(idist)
                  ENDDO
                  WRITE(io_unit, *)
               ENDIF
# 309 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
               IF (4 <= ndims) THEN
                  WRITE(io_unit, '(T7,A,1X)', advance='no') "Dist vec 4:"
                  DO idist = 1, SIZE(dist2_4)
                     WRITE(io_unit, '(I2,1X)', advance='no') dist2_4(idist)
                  ENDDO
                  WRITE(io_unit, *)
               ENDIF
# 317 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

               WRITE(io_unit,'(T4,A)') "Test distribution:"
# 320 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
               IF (1 <= ndims) THEN
                  WRITE(io_unit, '(T7,A,1X)', advance='no') "Dist vec 1:"
                  DO idist = 1, SIZE(dist2_1)
                     WRITE(io_unit, '(I2,1X)', advance='no') dist1_1(idist)
                  ENDDO
                  WRITE(io_unit, *)
               ENDIF
# 320 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
               IF (2 <= ndims) THEN
                  WRITE(io_unit, '(T7,A,1X)', advance='no') "Dist vec 2:"
                  DO idist = 1, SIZE(dist2_2)
                     WRITE(io_unit, '(I2,1X)', advance='no') dist1_2(idist)
                  ENDDO
                  WRITE(io_unit, *)
               ENDIF
# 320 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
               IF (3 <= ndims) THEN
                  WRITE(io_unit, '(T7,A,1X)', advance='no') "Dist vec 3:"
                  DO idist = 1, SIZE(dist2_3)
                     WRITE(io_unit, '(I2,1X)', advance='no') dist1_3(idist)
                  ENDDO
                  WRITE(io_unit, *)
               ENDIF
# 320 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
               IF (4 <= ndims) THEN
                  WRITE(io_unit, '(T7,A,1X)', advance='no') "Dist vec 4:"
                  DO idist = 1, SIZE(dist2_4)
                     WRITE(io_unit, '(I2,1X)', advance='no') dist1_4(idist)
                  ENDDO
                  WRITE(io_unit, *)
               ENDIF
# 328 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDIF

# 331 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF (ndims==2) THEN
               CALL dbcsr_t_distribution_new(dist2, mp_comm_ref, mp_comm_2d, map1_ref, map2_ref, dist2_1, dist2_2)
               CALL dbcsr_t_create(tensor2, "Ref", dist2, map1_ref, map2_ref, &
                                   dbcsr_type_real_8, blk_size_1, blk_size_2)
               CALL dbcsr_t_setup_test_tensor(tensor2, mp_comm_ref, blk_ind_1, blk_ind_2)
            ENDIF
# 331 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF (ndims==3) THEN
               CALL dbcsr_t_distribution_new(dist2, mp_comm_ref, mp_comm_2d, map1_ref, map2_ref, dist2_1, dist2_2, dist2_3)
               CALL dbcsr_t_create(tensor2, "Ref", dist2, map1_ref, map2_ref, &
                                   dbcsr_type_real_8, blk_size_1, blk_size_2, blk_size_3)
               CALL dbcsr_t_setup_test_tensor(tensor2, mp_comm_ref, blk_ind_1, blk_ind_2, blk_ind_3)
            ENDIF
# 331 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF (ndims==4) THEN
               CALL dbcsr_t_distribution_new(dist2, mp_comm_ref, mp_comm_2d, map1_ref, map2_ref, dist2_1, dist2_2, dist2_3, dist2_4)
               CALL dbcsr_t_create(tensor2, "Ref", dist2, map1_ref, map2_ref, &
                                   dbcsr_type_real_8, blk_size_1, blk_size_2, blk_size_3, blk_size_4)
               CALL dbcsr_t_setup_test_tensor(tensor2, mp_comm_ref, blk_ind_1, blk_ind_2, blk_ind_3, blk_ind_4)
            ENDIF
# 338 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

            IF(verbose) CALL write_tensor_blocks(tensor2, io_unit, unit_nr)

# 342 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF (ndims==2) THEN
               CALL dbcsr_t_distribution_new(dist1, mp_comm_test, mp_comm_2d, map1, map2, dist1_1, dist1_2)
               CALL dbcsr_t_create(tensor1, tensor_name, dist1, map1, map2, &
                                   dbcsr_type_real_8, blk_size_1, blk_size_2)
               CALL dbcsr_t_setup_test_tensor(tensor1, mp_comm_test, blk_ind_1, blk_ind_2)
            ENDIF
# 342 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF (ndims==3) THEN
               CALL dbcsr_t_distribution_new(dist1, mp_comm_test, mp_comm_2d, map1, map2, dist1_1, dist1_2, dist1_3)
               CALL dbcsr_t_create(tensor1, tensor_name, dist1, map1, map2, &
                                   dbcsr_type_real_8, blk_size_1, blk_size_2, blk_size_3)
               CALL dbcsr_t_setup_test_tensor(tensor1, mp_comm_test, blk_ind_1, blk_ind_2, blk_ind_3)
            ENDIF
# 342 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF (ndims==4) THEN
               CALL dbcsr_t_distribution_new(dist1, mp_comm_test, mp_comm_2d, map1, map2, dist1_1, dist1_2, dist1_3, dist1_4)
               CALL dbcsr_t_create(tensor1, tensor_name, dist1, map1, map2, &
                                   dbcsr_type_real_8, blk_size_1, blk_size_2, blk_size_3, blk_size_4)
               CALL dbcsr_t_setup_test_tensor(tensor1, mp_comm_test, blk_ind_1, blk_ind_2, blk_ind_3, blk_ind_4)
            ENDIF
# 349 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

            IF(verbose) CALL write_tensor_blocks(tensor1, io_unit, unit_nr)

            eql = dbcsr_t_equal(tensor1, tensor2)


            IF (.NOT. eql) THEN
               IF (io_unit > 0) WRITE(io_unit,'(A,1X,A)') TRIM(tensor_name), 'failed!'
               CPABORT('')
            ELSE
               IF (io_unit > 0) WRITE(io_unit,'(A,1X,A)') TRIM(tensor_name), 'passed!'
            ENDIF
            DEALLOCATE(map1, map2)

            CALL dbcsr_t_destroy(tensor1)
            CALL dbcsr_t_distribution_destroy(dist1)

            CALL mp_comm_free(mp_comm_test)
            CALL dbcsr_t_destroy(tensor2)
            CALL dbcsr_t_distribution_destroy(dist2)

         ENDDO
      ENDDO
      CALL mp_comm_free(mp_comm_ref)
   END SUBROUTINE

! **************************************************************************************************
!> \brief Create test distribution
!> \param dist_array ...
!> \param dist_size ...
!> \param nbins ...
!> \param mp_comm
! **************************************************************************************************
   SUBROUTINE dbcsr_t_random_dist(dist_array, dist_size, nbins, mp_comm)
      INTEGER, ALLOCATABLE, DIMENSION(:), INTENT(out) :: dist_array
      INTEGER, INTENT(IN)                             :: dist_size, nbins, mp_comm
      REAL, DIMENSION(dist_size)                      :: rn
      INTEGER                                         :: numnodes, mynode

      CALL mp_environ(numnodes, mynode, mp_comm)

      IF (mynode .EQ. 0) THEN
         CALL RANDOM_NUMBER(rn)
      ENDIF
      CALL mp_bcast(rn, 0, mp_comm)

      CALL allocate_any(dist_array, source=FLOOR(rn*nbins))

   END SUBROUTINE dbcsr_t_random_dist

! **************************************************************************************************
!> \brief Allocate and fill test tensor - entries are enumerated by their index s.t. they only depend
!>        on global properties of the tensor but not on distribution, matrix representation, etc.
!> \param tensor ...
!> \param mp_comm communicator
!> \param blk_ind_1 index along dimension 1 of non-zero blocks
! **************************************************************************************************
   SUBROUTINE dbcsr_t_setup_test_tensor(tensor, mp_comm, blk_ind_1, blk_ind_2, blk_ind_3, blk_ind_4)
      TYPE(dbcsr_t_type), INTENT(INOUT)                  :: tensor
      INTEGER, INTENT(IN)                                :: mp_comm
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: blk_ind_1, blk_ind_2, blk_ind_3, blk_ind_4
      INTEGER                                            :: blk, numnodes, mynode

      INTEGER                                            :: i, ib, my_nblks_alloc, nblks_alloc, proc
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: my_blk_ind_1, my_blk_ind_2, my_blk_ind_3, my_blk_ind_4
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))          :: blk_index, blk_offset, blk_size, &
                                                            tensor_dims
      INTEGER, DIMENSION(:,:), ALLOCATABLE               :: ind_nd
# 418 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), ALLOCATABLE, &
         DIMENSION(:,:)                :: blk_values_2
# 418 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), ALLOCATABLE, &
         DIMENSION(:,:,:)                :: blk_values_3
# 418 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), ALLOCATABLE, &
         DIMENSION(:,:,:,:)                :: blk_values_4
# 421 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      TYPE(dbcsr_t_iterator_type)                        :: iterator

      nblks_alloc = SIZE(blk_ind_1)
      CALL mp_environ(numnodes, mynode, mp_comm)

      ALLOCATE(ind_nd(nblks_alloc, dbcsr_t_ndims(tensor)))
      my_nblks_alloc = 0
      DO ib = 1, nblks_alloc
# 430 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(dbcsr_t_ndims(tensor) == 2) THEN
            ind_nd(ib,:) = [blk_ind_1(ib), blk_ind_2(ib)]
         ENDIF
# 430 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(dbcsr_t_ndims(tensor) == 3) THEN
            ind_nd(ib,:) = [blk_ind_1(ib), blk_ind_2(ib), blk_ind_3(ib)]
         ENDIF
# 430 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(dbcsr_t_ndims(tensor) == 4) THEN
            ind_nd(ib,:) = [blk_ind_1(ib), blk_ind_2(ib), blk_ind_3(ib), blk_ind_4(ib)]
         ENDIF
# 434 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         CALL dbcsr_t_get_stored_coordinates(tensor, ind_nd(ib,:), proc)
         IF (proc == mynode) THEN
            my_nblks_alloc = my_nblks_alloc+1
         ENDIF
      ENDDO

# 441 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      IF(dbcsr_t_ndims(tensor) >= 1) THEN
         ALLOCATE (my_blk_ind_1(my_nblks_alloc))
      ENDIF
# 441 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      IF(dbcsr_t_ndims(tensor) >= 2) THEN
         ALLOCATE (my_blk_ind_2(my_nblks_alloc))
      ENDIF
# 441 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      IF(dbcsr_t_ndims(tensor) >= 3) THEN
         ALLOCATE (my_blk_ind_3(my_nblks_alloc))
      ENDIF
# 441 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      IF(dbcsr_t_ndims(tensor) >= 4) THEN
         ALLOCATE (my_blk_ind_4(my_nblks_alloc))
      ENDIF
# 445 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

      i = 0
      DO ib = 1, nblks_alloc
         CALL dbcsr_t_get_stored_coordinates(tensor, ind_nd(ib,:), proc)
         IF (proc == mynode) THEN
            i = i+1
# 452 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF(dbcsr_t_ndims(tensor) >= 1) THEN
               my_blk_ind_1(i) = blk_ind_1(ib)
            ENDIF
# 452 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF(dbcsr_t_ndims(tensor) >= 2) THEN
               my_blk_ind_2(i) = blk_ind_2(ib)
            ENDIF
# 452 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF(dbcsr_t_ndims(tensor) >= 3) THEN
               my_blk_ind_3(i) = blk_ind_3(ib)
            ENDIF
# 452 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            IF(dbcsr_t_ndims(tensor) >= 4) THEN
               my_blk_ind_4(i) = blk_ind_4(ib)
            ENDIF
# 456 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         ENDIF
      ENDDO

# 460 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      IF(dbcsr_t_ndims(tensor) == 2) THEN
         CALL dbcsr_t_reserve_blocks(tensor, my_blk_ind_1, my_blk_ind_2)
      ENDIF
# 460 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      IF(dbcsr_t_ndims(tensor) == 3) THEN
         CALL dbcsr_t_reserve_blocks(tensor, my_blk_ind_1, my_blk_ind_2, my_blk_ind_3)
      ENDIF
# 460 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      IF(dbcsr_t_ndims(tensor) == 4) THEN
         CALL dbcsr_t_reserve_blocks(tensor, my_blk_ind_1, my_blk_ind_2, my_blk_ind_3, my_blk_ind_4)
      ENDIF
# 464 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, blk_index, blk, blk_size=blk_size, blk_offset=blk_offset)

# 470 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(dbcsr_t_ndims(tensor) == 2) THEN
            CALL allocate_any(blk_values_2, shape_spec=blk_size)
            CALL dims_tensor(tensor, tensor_dims)
            CALL enumerate_block_elements(blk_size, blk_offset, tensor_dims, blk_2=blk_values_2)
            CALL dbcsr_t_put_block(tensor, blk_index, blk_size, blk_values_2)
            DEALLOCATE (blk_values_2)
         ENDIF
# 470 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(dbcsr_t_ndims(tensor) == 3) THEN
            CALL allocate_any(blk_values_3, shape_spec=blk_size)
            CALL dims_tensor(tensor, tensor_dims)
            CALL enumerate_block_elements(blk_size, blk_offset, tensor_dims, blk_3=blk_values_3)
            CALL dbcsr_t_put_block(tensor, blk_index, blk_size, blk_values_3)
            DEALLOCATE (blk_values_3)
         ENDIF
# 470 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(dbcsr_t_ndims(tensor) == 4) THEN
            CALL allocate_any(blk_values_4, shape_spec=blk_size)
            CALL dims_tensor(tensor, tensor_dims)
            CALL enumerate_block_elements(blk_size, blk_offset, tensor_dims, blk_4=blk_values_4)
            CALL dbcsr_t_put_block(tensor, blk_index, blk_size, blk_values_4)
            DEALLOCATE (blk_values_4)
         ENDIF
# 478 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)

   END SUBROUTINE

! **************************************************************************************************
!> \brief Enumerate tensor entries in block
!> \param blk_size size of block
!> \param blk_offset block offset (indices of first element)
!> \param tensor_size global tensor sizes
!> \blk_2 block values for 2 dimensions
!> \blk_3 block values for 3 dimensions
!> \blk_4 ...
! **************************************************************************************************
   SUBROUTINE enumerate_block_elements(blk_size, blk_offset, tensor_size, blk_2, blk_3, blk_4)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: blk_size, blk_offset, tensor_size
# 495 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), DIMENSION(:,:), &
         OPTIONAL, INTENT(OUT)                           :: blk_2
# 495 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), DIMENSION(:,:,:), &
         OPTIONAL, INTENT(OUT)                           :: blk_3
# 495 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), DIMENSION(:,:,:,:), &
         OPTIONAL, INTENT(OUT)                           :: blk_4
# 498 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      INTEGER                                            :: ndim
      INTEGER, DIMENSION(SIZE(blk_size))                 :: arr_ind, tens_ind
      INTEGER                                            :: i_1, i_2, i_3, i_4

      ndim = SIZE(tensor_size)

# 505 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      IF (ndim == 2) THEN
# 507 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_2 = 1, blk_size(2)
# 507 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_1 = 1, blk_size(1)
# 509 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
               arr_ind(:) = [i_1, i_2]
               tens_ind(:) = arr_ind(:) + blk_offset(:) - 1
               blk_2(arr_ind(1), arr_ind(2)) = combine_index(tens_ind, tensor_size)
# 513 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 513 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 515 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDIF
# 505 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      IF (ndim == 3) THEN
# 507 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_3 = 1, blk_size(3)
# 507 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_2 = 1, blk_size(2)
# 507 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_1 = 1, blk_size(1)
# 509 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
               arr_ind(:) = [i_1, i_2, i_3]
               tens_ind(:) = arr_ind(:) + blk_offset(:) - 1
               blk_3(arr_ind(1), arr_ind(2), arr_ind(3)) = combine_index(tens_ind, tensor_size)
# 513 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 513 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 513 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 515 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDIF
# 505 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      IF (ndim == 4) THEN
# 507 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_4 = 1, blk_size(4)
# 507 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_3 = 1, blk_size(3)
# 507 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_2 = 1, blk_size(2)
# 507 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_1 = 1, blk_size(1)
# 509 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
               arr_ind(:) = [i_1, i_2, i_3, i_4]
               tens_ind(:) = arr_ind(:) + blk_offset(:) - 1
               blk_4(arr_ind(1), arr_ind(2), arr_ind(3), arr_ind(4)) = combine_index(tens_ind, tensor_size)
# 513 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 513 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 513 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 513 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 515 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDIF
# 517 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

   END SUBROUTINE

   SUBROUTINE write_tensor_block_indices(tensor, io_unit_master, io_unit_all)
      TYPE(dbcsr_t_type), INTENT(INOUT)                  :: tensor
      INTEGER, INTENT(IN)                                :: io_unit_master, io_unit_all
      TYPE(dbcsr_t_iterator_type)                        :: iterator
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))          :: blk_index, blk_size
      INTEGER                                            :: blk, mynode, numnodes, proc

      IF(io_unit_master > 0) THEN
         WRITE (io_unit_master, '(T7,A)') "(block index) @ process: size"
      ENDIF

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, blk_index, blk, blk_size=blk_size)
         CALL dbcsr_t_get_stored_coordinates(tensor, blk_index, proc)
         CALL mp_environ(numnodes, mynode, tensor%comm_nd)
         CPASSERT(proc .EQ. mynode)
# 538 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(dbcsr_t_ndims(tensor) == 2) THEN
            WRITE (io_unit_all, '(T7,A,T16,A,2I3,1X,A,1X,I3,A2,2I3)') &
               TRIM(tensor%name),"blk index (", blk_index,") @", proc,":", blk_size
         ENDIF
# 538 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(dbcsr_t_ndims(tensor) == 3) THEN
            WRITE (io_unit_all, '(T7,A,T16,A,3I3,1X,A,1X,I3,A2,3I3)') &
               TRIM(tensor%name),"blk index (", blk_index,") @", proc,":", blk_size
         ENDIF
# 538 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(dbcsr_t_ndims(tensor) == 4) THEN
            WRITE (io_unit_all, '(T7,A,T16,A,4I3,1X,A,1X,I3,A2,4I3)') &
               TRIM(tensor%name),"blk index (", blk_index,") @", proc,":", blk_size
         ENDIF
# 543 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
   END SUBROUTINE

! **************************************************************************************************
!> \brief Write all tensor blocks
!> \param tensor ...
!> \param io_unit_master for global output
!> \param io_unit_all for local output
! **************************************************************************************************
   SUBROUTINE write_tensor_blocks(tensor, io_unit_master, io_unit_all, write_int)
      TYPE(dbcsr_t_type), INTENT(INOUT)                  :: tensor
      INTEGER, INTENT(IN)                                :: io_unit_master, io_unit_all
      LOGICAL, INTENT(IN), OPTIONAL                      :: write_int
      INTEGER                                            :: blk
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))          :: blk_index, blk_size
# 560 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), ALLOCATABLE, &
         DIMENSION(:,:)                :: blk_values_2
# 560 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), ALLOCATABLE, &
         DIMENSION(:,:,:)                :: blk_values_3
# 560 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), ALLOCATABLE, &
         DIMENSION(:,:,:,:)                :: blk_values_4
# 563 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      TYPE(dbcsr_t_iterator_type)                        :: iterator
      INTEGER                                            :: proc, mynode, numnodes

      IF(io_unit_master > 0) THEN
         WRITE (io_unit_master, '(T7,A)') "(block index) @ process: (array index) value"
      ENDIF
      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, blk_index, blk, blk_size=blk_size)
         CALL dbcsr_t_get_stored_coordinates(tensor, blk_index, proc)
         CALL mp_environ(numnodes, mynode, tensor%comm_nd)
         CPASSERT(proc .EQ. mynode)
# 576 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(dbcsr_t_ndims(tensor) == 2) THEN
            CALL dbcsr_t_get_block(tensor, blk_index, blk_values_2)
            CALL write_block(tensor%name, blk_size, blk_index, proc, io_unit_all, &
                             blk_values_2=blk_values_2, write_int=write_int)
            DEALLOCATE (blk_values_2)
         ENDIF
# 576 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(dbcsr_t_ndims(tensor) == 3) THEN
            CALL dbcsr_t_get_block(tensor, blk_index, blk_values_3)
            CALL write_block(tensor%name, blk_size, blk_index, proc, io_unit_all, &
                             blk_values_3=blk_values_3, write_int=write_int)
            DEALLOCATE (blk_values_3)
         ENDIF
# 576 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(dbcsr_t_ndims(tensor) == 4) THEN
            CALL dbcsr_t_get_block(tensor, blk_index, blk_values_4)
            CALL write_block(tensor%name, blk_size, blk_index, proc, io_unit_all, &
                             blk_values_4=blk_values_4, write_int=write_int)
            DEALLOCATE (blk_values_4)
         ENDIF
# 583 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
   END SUBROUTINE

! **************************************************************************************************
!> \brief Write a tensor block
!> \param name tensor name
!> \param blk_size block size
!> \param blk_index block index
!> \param proc which process am I
!> \param unit_nr unit number
!> \param blk_values_2 block values for 2 dimensions
! **************************************************************************************************
   SUBROUTINE write_block(name, blk_size, blk_index, proc, unit_nr, &
                          blk_values_2, blk_values_3, blk_values_4, write_int)

      CHARACTER(LEN=*), INTENT(IN)                       :: name
      INTEGER, DIMENSION(:), INTENT(IN)                  :: blk_size, blk_index
# 602 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), &
         DIMENSION(blk_size(1), blk_size(2)), &
         INTENT(IN), OPTIONAL                            :: blk_values_2
# 602 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), &
         DIMENSION(blk_size(1), blk_size(2), blk_size(3)), &
         INTENT(IN), OPTIONAL                            :: blk_values_3
# 602 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), &
         DIMENSION(blk_size(1), blk_size(2), blk_size(3), blk_size(4)), &
         INTENT(IN), OPTIONAL                            :: blk_values_4
# 606 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      LOGICAL, INTENT(IN), OPTIONAL                      :: write_int
      LOGICAL                                            :: write_int_prv
      INTEGER, INTENT(IN)                                :: unit_nr, proc
      INTEGER                                            :: i_1, i_2, i_3, i_4
      INTEGER                                            :: ndim

      IF(PRESENT(write_int)) THEN
         write_int_prv = write_int
      ELSE
         write_int_prv = .FALSE.
      ENDIF

      ndim = SIZE(blk_size)

      IF (unit_nr > 0) THEN
# 622 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF (ndim == 2) THEN
# 624 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_2 = 1, blk_size(2)
# 624 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_1 = 1, blk_size(1)
# 626 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
               IF (write_int_prv) THEN
                  WRITE (unit_nr, '(T7,A,T16,A,2I3,1X,A,1X,I3,A,1X,A,2I3,1X,A,1X,I20)') &
                     TRIM(name),"(", blk_index,") @", proc, ':', &
                     "(",i_1, i_2,")", &
                     INT(blk_values_2(i_1, i_2), KIND=int_8)
               ELSE
                  WRITE (unit_nr, '(T7,A,T16,A,2I3,1X,A,1X,I3,A,1X,A,2I3,1X,A,1X,F10.5)') &
                     TRIM(name),"(", blk_index,") @", proc, ':', &
                     "(",i_1, i_2,")", &
                     blk_values_2(i_1, i_2)
               ENDIF
# 638 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 638 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 640 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         ENDIF
# 622 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF (ndim == 3) THEN
# 624 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_3 = 1, blk_size(3)
# 624 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_2 = 1, blk_size(2)
# 624 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_1 = 1, blk_size(1)
# 626 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
               IF (write_int_prv) THEN
                  WRITE (unit_nr, '(T7,A,T16,A,3I3,1X,A,1X,I3,A,1X,A,3I3,1X,A,1X,I20)') &
                     TRIM(name),"(", blk_index,") @", proc, ':', &
                     "(",i_1, i_2, i_3,")", &
                     INT(blk_values_3(i_1, i_2, i_3), KIND=int_8)
               ELSE
                  WRITE (unit_nr, '(T7,A,T16,A,3I3,1X,A,1X,I3,A,1X,A,3I3,1X,A,1X,F10.5)') &
                     TRIM(name),"(", blk_index,") @", proc, ':', &
                     "(",i_1, i_2, i_3,")", &
                     blk_values_3(i_1, i_2, i_3)
               ENDIF
# 638 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 638 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 638 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 640 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         ENDIF
# 622 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF (ndim == 4) THEN
# 624 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_4 = 1, blk_size(4)
# 624 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_3 = 1, blk_size(3)
# 624 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_2 = 1, blk_size(2)
# 624 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            DO i_1 = 1, blk_size(1)
# 626 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
               IF (write_int_prv) THEN
                  WRITE (unit_nr, '(T7,A,T16,A,4I3,1X,A,1X,I3,A,1X,A,4I3,1X,A,1X,I20)') &
                     TRIM(name),"(", blk_index,") @", proc, ':', &
                     "(",i_1, i_2, i_3, i_4,")", &
                     INT(blk_values_4(i_1, i_2, i_3, i_4), KIND=int_8)
               ELSE
                  WRITE (unit_nr, '(T7,A,T16,A,4I3,1X,A,1X,I3,A,1X,A,4I3,1X,A,1X,F10.5)') &
                     TRIM(name),"(", blk_index,") @", proc, ':', &
                     "(",i_1, i_2, i_3, i_4,")", &
                     blk_values_4(i_1, i_2, i_3, i_4)
               ENDIF
# 638 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 638 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 638 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 638 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
            ENDDO
# 640 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         ENDIF
# 642 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDIF
   END SUBROUTINE

# 646 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 647 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Transform a distributed sparse tensor to a replicated dense array. This is only useful for
!>        testing tensor contraction by matrix multiplication of dense arrays.
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE dist_sparse_tensor_to_repl_dense_2d_array_r_dp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                          :: tensor
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:),&
         INTENT(OUT)                                             :: array
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 2)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)
      CALL allocate_any(array, shape_spec=dims_nd)
      array(:,:) = 0.0_real_8

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL dbcsr_t_get_block(tensor, ind_nd, block)

         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         array(blk_start(1):blk_end(1), blk_start(2):blk_end(2)) = &
            block(:,:)

         DEALLOCATE(block)
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
      CALL mp_sum(array, tensor%comm_nd)

   END SUBROUTINE
# 647 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Transform a distributed sparse tensor to a replicated dense array. This is only useful for
!>        testing tensor contraction by matrix multiplication of dense arrays.
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE dist_sparse_tensor_to_repl_dense_3d_array_r_dp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                          :: tensor
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:),&
         INTENT(OUT)                                             :: array
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 3)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)
      CALL allocate_any(array, shape_spec=dims_nd)
      array(:,:,:) = 0.0_real_8

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL dbcsr_t_get_block(tensor, ind_nd, block)

         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3)) = &
            block(:,:,:)

         DEALLOCATE(block)
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
      CALL mp_sum(array, tensor%comm_nd)

   END SUBROUTINE
# 647 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Transform a distributed sparse tensor to a replicated dense array. This is only useful for
!>        testing tensor contraction by matrix multiplication of dense arrays.
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE dist_sparse_tensor_to_repl_dense_4d_array_r_dp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                          :: tensor
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:,:),&
         INTENT(OUT)                                             :: array
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 4)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)
      CALL allocate_any(array, shape_spec=dims_nd)
      array(:,:,:,:) = 0.0_real_8

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL dbcsr_t_get_block(tensor, ind_nd, block)

         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3), blk_start(4):blk_end(4)) = &
            block(:,:,:,:)

         DEALLOCATE(block)
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
      CALL mp_sum(array, tensor%comm_nd)

   END SUBROUTINE
# 688 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 646 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 647 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Transform a distributed sparse tensor to a replicated dense array. This is only useful for
!>        testing tensor contraction by matrix multiplication of dense arrays.
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE dist_sparse_tensor_to_repl_dense_2d_array_r_sp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                          :: tensor
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:),&
         INTENT(OUT)                                             :: array
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 2)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)
      CALL allocate_any(array, shape_spec=dims_nd)
      array(:,:) = 0.0_real_4

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL dbcsr_t_get_block(tensor, ind_nd, block)

         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         array(blk_start(1):blk_end(1), blk_start(2):blk_end(2)) = &
            block(:,:)

         DEALLOCATE(block)
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
      CALL mp_sum(array, tensor%comm_nd)

   END SUBROUTINE
# 647 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Transform a distributed sparse tensor to a replicated dense array. This is only useful for
!>        testing tensor contraction by matrix multiplication of dense arrays.
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE dist_sparse_tensor_to_repl_dense_3d_array_r_sp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                          :: tensor
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:),&
         INTENT(OUT)                                             :: array
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 3)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)
      CALL allocate_any(array, shape_spec=dims_nd)
      array(:,:,:) = 0.0_real_4

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL dbcsr_t_get_block(tensor, ind_nd, block)

         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3)) = &
            block(:,:,:)

         DEALLOCATE(block)
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
      CALL mp_sum(array, tensor%comm_nd)

   END SUBROUTINE
# 647 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Transform a distributed sparse tensor to a replicated dense array. This is only useful for
!>        testing tensor contraction by matrix multiplication of dense arrays.
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE dist_sparse_tensor_to_repl_dense_4d_array_r_sp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                          :: tensor
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:,:),&
         INTENT(OUT)                                             :: array
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 4)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)
      CALL allocate_any(array, shape_spec=dims_nd)
      array(:,:,:,:) = 0.0_real_4

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL dbcsr_t_get_block(tensor, ind_nd, block)

         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3), blk_start(4):blk_end(4)) = &
            block(:,:,:,:)

         DEALLOCATE(block)
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
      CALL mp_sum(array, tensor%comm_nd)

   END SUBROUTINE
# 688 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 646 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 647 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Transform a distributed sparse tensor to a replicated dense array. This is only useful for
!>        testing tensor contraction by matrix multiplication of dense arrays.
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE dist_sparse_tensor_to_repl_dense_2d_array_c_dp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                          :: tensor
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:),&
         INTENT(OUT)                                             :: array
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 2)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)
      CALL allocate_any(array, shape_spec=dims_nd)
      array(:,:) = 0.0_real_8

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL dbcsr_t_get_block(tensor, ind_nd, block)

         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         array(blk_start(1):blk_end(1), blk_start(2):blk_end(2)) = &
            block(:,:)

         DEALLOCATE(block)
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
      CALL mp_sum(array, tensor%comm_nd)

   END SUBROUTINE
# 647 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Transform a distributed sparse tensor to a replicated dense array. This is only useful for
!>        testing tensor contraction by matrix multiplication of dense arrays.
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE dist_sparse_tensor_to_repl_dense_3d_array_c_dp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                          :: tensor
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:),&
         INTENT(OUT)                                             :: array
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 3)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)
      CALL allocate_any(array, shape_spec=dims_nd)
      array(:,:,:) = 0.0_real_8

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL dbcsr_t_get_block(tensor, ind_nd, block)

         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3)) = &
            block(:,:,:)

         DEALLOCATE(block)
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
      CALL mp_sum(array, tensor%comm_nd)

   END SUBROUTINE
# 647 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Transform a distributed sparse tensor to a replicated dense array. This is only useful for
!>        testing tensor contraction by matrix multiplication of dense arrays.
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE dist_sparse_tensor_to_repl_dense_4d_array_c_dp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                          :: tensor
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:,:),&
         INTENT(OUT)                                             :: array
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 4)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)
      CALL allocate_any(array, shape_spec=dims_nd)
      array(:,:,:,:) = 0.0_real_8

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL dbcsr_t_get_block(tensor, ind_nd, block)

         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3), blk_start(4):blk_end(4)) = &
            block(:,:,:,:)

         DEALLOCATE(block)
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
      CALL mp_sum(array, tensor%comm_nd)

   END SUBROUTINE
# 688 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 646 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 647 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Transform a distributed sparse tensor to a replicated dense array. This is only useful for
!>        testing tensor contraction by matrix multiplication of dense arrays.
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE dist_sparse_tensor_to_repl_dense_2d_array_c_sp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                          :: tensor
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:),&
         INTENT(OUT)                                             :: array
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 2)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)
      CALL allocate_any(array, shape_spec=dims_nd)
      array(:,:) = 0.0_real_4

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL dbcsr_t_get_block(tensor, ind_nd, block)

         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         array(blk_start(1):blk_end(1), blk_start(2):blk_end(2)) = &
            block(:,:)

         DEALLOCATE(block)
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
      CALL mp_sum(array, tensor%comm_nd)

   END SUBROUTINE
# 647 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Transform a distributed sparse tensor to a replicated dense array. This is only useful for
!>        testing tensor contraction by matrix multiplication of dense arrays.
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE dist_sparse_tensor_to_repl_dense_3d_array_c_sp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                          :: tensor
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:),&
         INTENT(OUT)                                             :: array
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 3)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)
      CALL allocate_any(array, shape_spec=dims_nd)
      array(:,:,:) = 0.0_real_4

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL dbcsr_t_get_block(tensor, ind_nd, block)

         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3)) = &
            block(:,:,:)

         DEALLOCATE(block)
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
      CALL mp_sum(array, tensor%comm_nd)

   END SUBROUTINE
# 647 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Transform a distributed sparse tensor to a replicated dense array. This is only useful for
!>        testing tensor contraction by matrix multiplication of dense arrays.
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE dist_sparse_tensor_to_repl_dense_4d_array_c_sp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                          :: tensor
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:,:),&
         INTENT(OUT)                                             :: array
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 4)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)
      CALL allocate_any(array, shape_spec=dims_nd)
      array(:,:,:,:) = 0.0_real_4

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL dbcsr_t_get_block(tensor, ind_nd, block)

         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3), blk_start(4):blk_end(4)) = &
            block(:,:,:,:)

         DEALLOCATE(block)
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
      CALL mp_sum(array, tensor%comm_nd)

   END SUBROUTINE
# 688 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 689 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

# 691 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 692 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief ...
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE repl_dense_2d_array_to_dist_sparse_tensor_r_dp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                             :: tensor
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:),&
         INTENT(INOUT)                                           :: array
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 2)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL allocate_any(block, shape_spec=blk_size)
         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         block(:,:) = &
            array(blk_start(1):blk_end(1), blk_start(2):blk_end(2))
         CALL dbcsr_t_put_block(tensor, ind_nd, blk_size, block)
         DEALLOCATE(block)
      ENDDO

   END SUBROUTINE
# 692 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief ...
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE repl_dense_3d_array_to_dist_sparse_tensor_r_dp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                             :: tensor
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:),&
         INTENT(INOUT)                                           :: array
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 3)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL allocate_any(block, shape_spec=blk_size)
         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         block(:,:,:) = &
            array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3))
         CALL dbcsr_t_put_block(tensor, ind_nd, blk_size, block)
         DEALLOCATE(block)
      ENDDO

   END SUBROUTINE
# 692 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief ...
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE repl_dense_4d_array_to_dist_sparse_tensor_r_dp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                             :: tensor
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:,:),&
         INTENT(INOUT)                                           :: array
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 4)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL allocate_any(block, shape_spec=blk_size)
         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         block(:,:,:,:) = &
            array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3), blk_start(4):blk_end(4))
         CALL dbcsr_t_put_block(tensor, ind_nd, blk_size, block)
         DEALLOCATE(block)
      ENDDO

   END SUBROUTINE
# 727 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 691 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 692 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief ...
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE repl_dense_2d_array_to_dist_sparse_tensor_r_sp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                             :: tensor
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:),&
         INTENT(INOUT)                                           :: array
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 2)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL allocate_any(block, shape_spec=blk_size)
         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         block(:,:) = &
            array(blk_start(1):blk_end(1), blk_start(2):blk_end(2))
         CALL dbcsr_t_put_block(tensor, ind_nd, blk_size, block)
         DEALLOCATE(block)
      ENDDO

   END SUBROUTINE
# 692 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief ...
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE repl_dense_3d_array_to_dist_sparse_tensor_r_sp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                             :: tensor
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:),&
         INTENT(INOUT)                                           :: array
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 3)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL allocate_any(block, shape_spec=blk_size)
         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         block(:,:,:) = &
            array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3))
         CALL dbcsr_t_put_block(tensor, ind_nd, blk_size, block)
         DEALLOCATE(block)
      ENDDO

   END SUBROUTINE
# 692 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief ...
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE repl_dense_4d_array_to_dist_sparse_tensor_r_sp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                             :: tensor
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:,:),&
         INTENT(INOUT)                                           :: array
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 4)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL allocate_any(block, shape_spec=blk_size)
         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         block(:,:,:,:) = &
            array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3), blk_start(4):blk_end(4))
         CALL dbcsr_t_put_block(tensor, ind_nd, blk_size, block)
         DEALLOCATE(block)
      ENDDO

   END SUBROUTINE
# 727 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 691 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 692 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief ...
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE repl_dense_2d_array_to_dist_sparse_tensor_c_dp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                             :: tensor
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:),&
         INTENT(INOUT)                                           :: array
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 2)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL allocate_any(block, shape_spec=blk_size)
         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         block(:,:) = &
            array(blk_start(1):blk_end(1), blk_start(2):blk_end(2))
         CALL dbcsr_t_put_block(tensor, ind_nd, blk_size, block)
         DEALLOCATE(block)
      ENDDO

   END SUBROUTINE
# 692 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief ...
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE repl_dense_3d_array_to_dist_sparse_tensor_c_dp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                             :: tensor
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:),&
         INTENT(INOUT)                                           :: array
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 3)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL allocate_any(block, shape_spec=blk_size)
         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         block(:,:,:) = &
            array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3))
         CALL dbcsr_t_put_block(tensor, ind_nd, blk_size, block)
         DEALLOCATE(block)
      ENDDO

   END SUBROUTINE
# 692 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief ...
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE repl_dense_4d_array_to_dist_sparse_tensor_c_dp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                             :: tensor
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:,:),&
         INTENT(INOUT)                                           :: array
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 4)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL allocate_any(block, shape_spec=blk_size)
         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         block(:,:,:,:) = &
            array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3), blk_start(4):blk_end(4))
         CALL dbcsr_t_put_block(tensor, ind_nd, blk_size, block)
         DEALLOCATE(block)
      ENDDO

   END SUBROUTINE
# 727 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 691 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 692 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief ...
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE repl_dense_2d_array_to_dist_sparse_tensor_c_sp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                             :: tensor
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:),&
         INTENT(INOUT)                                           :: array
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 2)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL allocate_any(block, shape_spec=blk_size)
         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         block(:,:) = &
            array(blk_start(1):blk_end(1), blk_start(2):blk_end(2))
         CALL dbcsr_t_put_block(tensor, ind_nd, blk_size, block)
         DEALLOCATE(block)
      ENDDO

   END SUBROUTINE
# 692 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief ...
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE repl_dense_3d_array_to_dist_sparse_tensor_c_sp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                             :: tensor
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:),&
         INTENT(INOUT)                                           :: array
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 3)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL allocate_any(block, shape_spec=blk_size)
         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         block(:,:,:) = &
            array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3))
         CALL dbcsr_t_put_block(tensor, ind_nd, blk_size, block)
         DEALLOCATE(block)
      ENDDO

   END SUBROUTINE
# 692 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief ...
!> \param tensor
!> \param array
! **************************************************************************************************
   SUBROUTINE repl_dense_4d_array_to_dist_sparse_tensor_c_sp(tensor, array)
      TYPE(dbcsr_t_type), INTENT(INOUT)                             :: tensor
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:,:),&
         INTENT(INOUT)                                           :: array
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:,:)   :: block
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: dims_nd, ind_nd, blk_size, blk_offset
      TYPE(dbcsr_t_iterator_type)                                     :: iterator
      INTEGER                                                    :: blk, idim
      INTEGER, DIMENSION(dbcsr_t_ndims(tensor))                  :: blk_start, blk_end


      CPASSERT(dbcsr_t_ndims(tensor) .EQ. 4)
      CALL get_mapping_info(tensor%nd_index, dims_nd=dims_nd)

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_size=blk_size, blk_offset=blk_offset)
         CALL allocate_any(block, shape_spec=blk_size)
         DO idim = 1, dbcsr_t_ndims(tensor)
            blk_start(idim) = blk_offset(idim)
            blk_end(idim) = blk_offset(idim)+blk_size(idim) - 1
         ENDDO
         block(:,:,:,:) = &
            array(blk_start(1):blk_end(1), blk_start(2):blk_end(2), blk_start(3):blk_end(3), blk_start(4):blk_end(4))
         CALL dbcsr_t_put_block(tensor, ind_nd, blk_size, block)
         DEALLOCATE(block)
      ENDDO

   END SUBROUTINE
# 727 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
# 728 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

! **************************************************************************************************
!> \brief test tensor contraction
!> \note for testing/debugging, simply replace a call to dbcsr_t_contract with a call to this routine
!> \param tensor_1 ...
!> \param tensor_2 ...
!> \param contract_1 ...
!> \param notcontract_1 ...
!> \param contract_2 ...
!> \param notcontract_2 ...
!> \param map_1 ...
!> \param map_2 ...
!> \param unit_nr
! **************************************************************************************************
   SUBROUTINE dbcsr_t_contract_test(tensor_1, tensor_2, tensor_3, &
                                    contract_1, notcontract_1, &
                                    contract_2, notcontract_2, &
                                    map_1, map_2, unit_nr, write_int)
      TYPE(dbcsr_t_type), INTENT(INOUT)    :: tensor_1, tensor_2, tensor_3
      INTEGER, DIMENSION(:), INTENT(IN)    :: contract_1, contract_2, &
                                              notcontract_1, notcontract_2,&
                                              map_1, map_2
      INTEGER, INTENT(IN)                  :: unit_nr
      LOGICAL, INTENT(IN), OPTIONAL        :: write_int
      INTEGER                              :: io_unit, mynode, numnodes, mp_comm
      INTEGER, DIMENSION(:), ALLOCATABLE   :: size_1, size_2, size_3, &
                                              order_t1, order_t2, order_t3

# 757 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), ALLOCATABLE, &
         DIMENSION(:,:) :: array_1_2d, &
                                             array_2_2d, &
                                             array_3_2d, &
                                             array_1_rs2d, &
                                             array_2_rs2d, &
                                             array_3_rs2d
# 757 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), ALLOCATABLE, &
         DIMENSION(:,:,:) :: array_1_3d, &
                                             array_2_3d, &
                                             array_3_3d, &
                                             array_1_rs3d, &
                                             array_2_rs3d, &
                                             array_3_rs3d
# 757 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), ALLOCATABLE, &
         DIMENSION(:,:,:,:) :: array_1_4d, &
                                             array_2_4d, &
                                             array_3_4d, &
                                             array_1_rs4d, &
                                             array_2_rs4d, &
                                             array_3_rs4d
# 765 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      REAL(KIND=real_8), ALLOCATABLE, &
         DIMENSION(:,:)                   :: array_1_mm, &
                                             array_2_mm, &
                                             array_3_mm, &
                                             array_3_test_mm
      LOGICAL                             :: eql
      LOGICAL, PARAMETER                  :: debug = .FALSE.

      CALL dbcsr_get_info(tensor_1%matrix_rep, group=mp_comm)
      CALL mp_environ(numnodes, mynode, mp_comm)
      io_unit = 0
      IF (mynode .EQ. 0) io_unit = unit_nr

      IF (debug) THEN
         CALL write_tensor_block_indices(tensor_1, io_unit, unit_nr)
         CALL write_tensor_blocks(tensor_1, io_unit, unit_nr, write_int)
      ENDIF

      IF(io_unit > 0) THEN
         WRITE(io_unit, *)
         WRITE(io_unit, '(A)') repeat("-",80)
         WRITE(io_unit, '(A,1X,A,1X,A,1X,A,1X,A,1X,A)') "Testing tensor contraction", &
            TRIM(tensor_1%name), "x", TRIM(tensor_2%name), "=", TRIM(tensor_3%name)
         WRITE(io_unit, '(A)') repeat("-",80)
      ENDIF

      CALL dbcsr_t_contract(tensor_1, tensor_2, tensor_3, &
                            contract_1, notcontract_1, &
                            contract_2, notcontract_2, &
                            map_1, map_2)

      ! Convert tensors to simple multidimensional arrays
# 798 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      SELECT CASE(dbcsr_t_ndims(tensor_1))
# 800 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(2)
         CALL dist_sparse_tensor_to_repl_dense_array(tensor_1, array_1_2d)

# 800 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(3)
         CALL dist_sparse_tensor_to_repl_dense_array(tensor_1, array_1_3d)

# 800 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(4)
         CALL dist_sparse_tensor_to_repl_dense_array(tensor_1, array_1_4d)

# 804 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      END SELECT
# 798 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      SELECT CASE(dbcsr_t_ndims(tensor_2))
# 800 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(2)
         CALL dist_sparse_tensor_to_repl_dense_array(tensor_2, array_2_2d)

# 800 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(3)
         CALL dist_sparse_tensor_to_repl_dense_array(tensor_2, array_2_3d)

# 800 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(4)
         CALL dist_sparse_tensor_to_repl_dense_array(tensor_2, array_2_4d)

# 804 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      END SELECT
# 798 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      SELECT CASE(dbcsr_t_ndims(tensor_3))
# 800 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(2)
         CALL dist_sparse_tensor_to_repl_dense_array(tensor_3, array_3_2d)

# 800 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(3)
         CALL dist_sparse_tensor_to_repl_dense_array(tensor_3, array_3_3d)

# 800 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(4)
         CALL dist_sparse_tensor_to_repl_dense_array(tensor_3, array_3_4d)

# 804 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      END SELECT
# 806 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

      ! Get array sizes

# 810 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      SELECT CASE(dbcsr_t_ndims(tensor_1))
# 812 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(2)
         CALL allocate_any(size_1, source=SHAPE(array_1_2d))

# 812 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(3)
         CALL allocate_any(size_1, source=SHAPE(array_1_3d))

# 812 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(4)
         CALL allocate_any(size_1, source=SHAPE(array_1_4d))

# 816 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      END SELECT
# 810 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      SELECT CASE(dbcsr_t_ndims(tensor_2))
# 812 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(2)
         CALL allocate_any(size_2, source=SHAPE(array_2_2d))

# 812 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(3)
         CALL allocate_any(size_2, source=SHAPE(array_2_3d))

# 812 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(4)
         CALL allocate_any(size_2, source=SHAPE(array_2_4d))

# 816 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      END SELECT
# 810 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      SELECT CASE(dbcsr_t_ndims(tensor_3))
# 812 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(2)
         CALL allocate_any(size_3, source=SHAPE(array_3_2d))

# 812 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(3)
         CALL allocate_any(size_3, source=SHAPE(array_3_3d))

# 812 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(4)
         CALL allocate_any(size_3, source=SHAPE(array_3_4d))

# 816 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      END SELECT
# 818 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

# 820 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ALLOCATE(order_t1(dbcsr_t_ndims(tensor_1)))
# 820 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ALLOCATE(order_t2(dbcsr_t_ndims(tensor_2)))
# 820 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ALLOCATE(order_t3(dbcsr_t_ndims(tensor_3)))
# 822 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"


      ASSOCIATE(map_t1_1=>notcontract_1, map_t1_2=>contract_1, &
                map_t2_1=>notcontract_2, map_t2_2=>contract_2, &
                map_t3_1=>map_1, map_t3_2=>map_2)

# 829 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      order_t1(:) = dbcsr_t_inverse_order([map_t1_1, map_t1_2])

      SELECT CASE(dbcsr_t_ndims(tensor_1))
# 833 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(2)
         CALL allocate_any(array_1_rs2d, source=array_1_2d, order=order_t1)
         CALL allocate_any(array_1_mm, sizes_2d(size_1, map_t1_1, map_t1_2))
         array_1_mm(:,:) = RESHAPE(array_1_rs2d, SHAPE(array_1_mm))
# 833 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(3)
         CALL allocate_any(array_1_rs3d, source=array_1_3d, order=order_t1)
         CALL allocate_any(array_1_mm, sizes_2d(size_1, map_t1_1, map_t1_2))
         array_1_mm(:,:) = RESHAPE(array_1_rs3d, SHAPE(array_1_mm))
# 833 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(4)
         CALL allocate_any(array_1_rs4d, source=array_1_4d, order=order_t1)
         CALL allocate_any(array_1_mm, sizes_2d(size_1, map_t1_1, map_t1_2))
         array_1_mm(:,:) = RESHAPE(array_1_rs4d, SHAPE(array_1_mm))
# 838 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      END SELECT
# 829 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      order_t2(:) = dbcsr_t_inverse_order([map_t2_1, map_t2_2])

      SELECT CASE(dbcsr_t_ndims(tensor_2))
# 833 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(2)
         CALL allocate_any(array_2_rs2d, source=array_2_2d, order=order_t2)
         CALL allocate_any(array_2_mm, sizes_2d(size_2, map_t2_1, map_t2_2))
         array_2_mm(:,:) = RESHAPE(array_2_rs2d, SHAPE(array_2_mm))
# 833 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(3)
         CALL allocate_any(array_2_rs3d, source=array_2_3d, order=order_t2)
         CALL allocate_any(array_2_mm, sizes_2d(size_2, map_t2_1, map_t2_2))
         array_2_mm(:,:) = RESHAPE(array_2_rs3d, SHAPE(array_2_mm))
# 833 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(4)
         CALL allocate_any(array_2_rs4d, source=array_2_4d, order=order_t2)
         CALL allocate_any(array_2_mm, sizes_2d(size_2, map_t2_1, map_t2_2))
         array_2_mm(:,:) = RESHAPE(array_2_rs4d, SHAPE(array_2_mm))
# 838 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      END SELECT
# 829 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      order_t3(:) = dbcsr_t_inverse_order([map_t3_1, map_t3_2])

      SELECT CASE(dbcsr_t_ndims(tensor_3))
# 833 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(2)
         CALL allocate_any(array_3_rs2d, source=array_3_2d, order=order_t3)
         CALL allocate_any(array_3_mm, sizes_2d(size_3, map_t3_1, map_t3_2))
         array_3_mm(:,:) = RESHAPE(array_3_rs2d, SHAPE(array_3_mm))
# 833 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(3)
         CALL allocate_any(array_3_rs3d, source=array_3_3d, order=order_t3)
         CALL allocate_any(array_3_mm, sizes_2d(size_3, map_t3_1, map_t3_2))
         array_3_mm(:,:) = RESHAPE(array_3_rs3d, SHAPE(array_3_mm))
# 833 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      CASE(4)
         CALL allocate_any(array_3_rs4d, source=array_3_4d, order=order_t3)
         CALL allocate_any(array_3_mm, sizes_2d(size_3, map_t3_1, map_t3_2))
         array_3_mm(:,:) = RESHAPE(array_3_rs4d, SHAPE(array_3_mm))
# 838 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      END SELECT
# 840 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

      CALL allocate_any(array_3_test_mm, shape_spec=SHAPE(array_3_mm))
      array_3_test_mm(:,:) = MATMUL(array_1_mm, transpose(array_2_mm))

      END ASSOCIATE

      eql = MAXVAL(ABS(array_3_test_mm(:,:) - array_3_mm(:,:))) .LT. 1.0E-12_real_4

      IF (.NOT. eql) THEN
         IF (io_unit > 0) WRITE(io_unit,'(A)') 'Test failed!'
         CPABORT('')
      ELSE
         IF (io_unit > 0) WRITE(io_unit,'(A)') 'Test passed!'
      ENDIF

   END SUBROUTINE

# 858 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Write nd array
!> \param array ...
!> \param unit_nr ...
!> \param array_compare ...
! **************************************************************************************************
   SUBROUTINE write_2d_array(array, unit_nr, array_compare)
      REAL(KIND=real_8), DIMENSION(:,:), INTENT(IN) :: array
      INTEGER, INTENT(IN)                                             :: unit_nr
      REAL(KIND=real_8), DIMENSION(:,:), INTENT(IN), OPTIONAL :: array_compare
      INTEGER                                            :: i_1, i_2
# 870 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      DO i_2 = 1, SIZE(array, 2)
# 870 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      DO i_1 = 1, SIZE(array, 1)
# 872 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(INT(array(i_1, i_2), KIND=int_8) .GT. 0) THEN
            WRITE(unit_nr, '(3I20)', advance='no') i_1, i_2, INT(array(i_1, i_2), &
               KIND=int_8)
            IF(PRESENT(array_compare)) THEN
               WRITE(unit_nr, '(I20)', advance='no') INT(array_compare(i_1, i_2), KIND=int_8)
            ENDIF
            WRITE(unit_nr,*)
         ENDIF
# 881 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDDO
# 881 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDDO
# 883 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
   END SUBROUTINE
# 858 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Write nd array
!> \param array ...
!> \param unit_nr ...
!> \param array_compare ...
! **************************************************************************************************
   SUBROUTINE write_3d_array(array, unit_nr, array_compare)
      REAL(KIND=real_8), DIMENSION(:,:,:), INTENT(IN) :: array
      INTEGER, INTENT(IN)                                             :: unit_nr
      REAL(KIND=real_8), DIMENSION(:,:,:), INTENT(IN), OPTIONAL :: array_compare
      INTEGER                                            :: i_1, i_2, i_3
# 870 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      DO i_3 = 1, SIZE(array, 3)
# 870 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      DO i_2 = 1, SIZE(array, 2)
# 870 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      DO i_1 = 1, SIZE(array, 1)
# 872 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(INT(array(i_1, i_2, i_3), KIND=int_8) .GT. 0) THEN
            WRITE(unit_nr, '(4I20)', advance='no') i_1, i_2, i_3, INT(array(i_1, i_2, i_3), &
               KIND=int_8)
            IF(PRESENT(array_compare)) THEN
               WRITE(unit_nr, '(I20)', advance='no') INT(array_compare(i_1, i_2, i_3), KIND=int_8)
            ENDIF
            WRITE(unit_nr,*)
         ENDIF
# 881 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDDO
# 881 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDDO
# 881 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDDO
# 883 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
   END SUBROUTINE
# 858 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
! **************************************************************************************************
!> \brief Write nd array
!> \param array ...
!> \param unit_nr ...
!> \param array_compare ...
! **************************************************************************************************
   SUBROUTINE write_4d_array(array, unit_nr, array_compare)
      REAL(KIND=real_8), DIMENSION(:,:,:,:), INTENT(IN) :: array
      INTEGER, INTENT(IN)                                             :: unit_nr
      REAL(KIND=real_8), DIMENSION(:,:,:,:), INTENT(IN), OPTIONAL :: array_compare
      INTEGER                                            :: i_1, i_2, i_3, i_4
# 870 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      DO i_4 = 1, SIZE(array, 4)
# 870 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      DO i_3 = 1, SIZE(array, 3)
# 870 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      DO i_2 = 1, SIZE(array, 2)
# 870 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      DO i_1 = 1, SIZE(array, 1)
# 872 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
         IF(INT(array(i_1, i_2, i_3, i_4), KIND=int_8) .GT. 0) THEN
            WRITE(unit_nr, '(5I20)', advance='no') i_1, i_2, i_3, i_4, INT(array(i_1, i_2, i_3, i_4), &
               KIND=int_8)
            IF(PRESENT(array_compare)) THEN
               WRITE(unit_nr, '(I20)', advance='no') INT(array_compare(i_1, i_2, i_3, i_4), KIND=int_8)
            ENDIF
            WRITE(unit_nr,*)
         ENDIF
# 881 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDDO
# 881 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDDO
# 881 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDDO
# 881 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
      ENDDO
# 883 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"
   END SUBROUTINE
# 885 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_test.F"

! **************************************************************************************************
!> \brief mapped sizes in 2d
!> \param nd_sizes ...
!> \param map1 ...
!> \param map2 ...
! **************************************************************************************************
   FUNCTION sizes_2d(nd_sizes, map1, map2)
      INTEGER, DIMENSION(:), INTENT(IN) :: nd_sizes, map1, map2
      INTEGER, DIMENSION(2)             :: sizes_2d
      sizes_2d(1) = PRODUCT(nd_sizes(map1))
      sizes_2d(2) = PRODUCT(nd_sizes(map2))
   END FUNCTION

END MODULE
