# 1 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2018  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utility routines for the memory handling.
!> \par History
!>      (12.2017) remove stop_memory
!> \author Matthias Krack (25.06.1999)
! **************************************************************************************************
MODULE memory_utilities

   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp, int_8
#include "../base/base_uses.f90"

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'memory_utilities'

   PUBLIC :: reallocate

   INTERFACE reallocate
      MODULE PROCEDURE reallocate_c1, reallocate_c2, reallocate_c3, reallocate_c4, &
         reallocate_i1, reallocate_i2, reallocate_i3, reallocate_i4, &
         reallocate_r1, reallocate_r2, reallocate_r3, reallocate_r4, &
         reallocate_r5, reallocate_s1, reallocate_l1, reallocate_8i1, &
         reallocate_8i2
   END INTERFACE

CONTAINS

# 41 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
# 76 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

! **************************************************************************************************
!> \brief (Re)Allocate a complex vector with a new dimension
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \par History
!>      none
!> \author Matthias Krack (28.11.2005,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_c1(p, lb1_new, ub1_new)

      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_c1', routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: zero = (0.0_dp, 0.0_dp)

      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: work


# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1 ))
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1) = p(lb1:ub1)
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new))
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1) = work(lb1:ub1)
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 96 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_c1

! **************************************************************************************************
!> \brief (Re)Allocate a complex array 2D with a new dimension
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \par History
!>      none
!> \author Matthias Krack (28.11.2005,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_c2(p, lb1_new, ub1_new, lb2_new, ub2_new)

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_c2', routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: zero = (0.0_dp, 0.0_dp)

      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :)     :: work


# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb2, lb2_old, ub2, ub2_old
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2_old = LBOUND(p,2)
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2_old = UBOUND(p,2)
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2 = MAX(lb2_new,lb2_old)
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2 = MIN(ub2_new,ub2_old)
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1,lb2:ub2 ))
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1,lb2:ub2) = p(lb1:ub1,lb2:ub2)
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new))
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1,lb2:ub2) = work(lb1:ub1,lb2:ub2)
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 121 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_c2

! **************************************************************************************************
!> \brief (Re)Allocate a complex array 3D with a new dimension
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \param lb3_new ...
!> \param ub3_new ...
!> \par History
!>      none
!> \author Matthias Krack (28.11.2005,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_c3(p, lb1_new, ub1_new, lb2_new, ub2_new, lb3_new, ub3_new)

      COMPLEX(KIND=dp), DIMENSION(:, :, :), POINTER      :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new, &
                                                            lb3_new, ub3_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_c3', routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: zero = (0.0_dp, 0.0_dp)

      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)  :: work


# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb2, lb2_old, ub2, ub2_old
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb3, lb3_old, ub3, ub3_old
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2_old = LBOUND(p,2)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2_old = UBOUND(p,2)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2 = MAX(lb2_new,lb2_old)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2 = MIN(ub2_new,ub2_old)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb3_old = LBOUND(p,3)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub3_old = UBOUND(p,3)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb3 = MAX(lb3_new,lb3_old)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub3 = MIN(ub3_new,ub3_old)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1,lb2:ub2,lb3:ub3 ))
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1,lb2:ub2,lb3:ub3) = p(lb1:ub1,lb2:ub2,lb3:ub3)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new,lb3_new:ub3_new))
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1,lb2:ub2,lb3:ub3) = work(lb1:ub1,lb2:ub2,lb3:ub3)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 149 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_c3

! **************************************************************************************************
!> \brief (Re)Allocate a complex array 4D with a new dimension
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \param lb3_new ...
!> \param ub3_new ...
!> \param lb4_new ...
!> \param ub4_new ...
!> \par History
!>      none
!> \author Matthias Krack (28.11.2005,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_c4(p, lb1_new, ub1_new, lb2_new, ub2_new, lb3_new, ub3_new, &
                            lb4_new, ub4_new)

      COMPLEX(KIND=dp), DIMENSION(:, :, :, :), POINTER   :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new, &
                                                            lb3_new, ub3_new, lb4_new, ub4_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_c4', routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: zero = (0.0_dp, 0.0_dp)

      COMPLEX(KIND=dp), ALLOCATABLE, &
         DIMENSION(:, :, :, :)                           :: work


# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb2, lb2_old, ub2, ub2_old
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb3, lb3_old, ub3, ub3_old
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb4, lb4_old, ub4, ub4_old
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2_old = LBOUND(p,2)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2_old = UBOUND(p,2)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2 = MAX(lb2_new,lb2_old)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2 = MIN(ub2_new,ub2_old)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb3_old = LBOUND(p,3)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub3_old = UBOUND(p,3)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb3 = MAX(lb3_new,lb3_old)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub3 = MIN(ub3_new,ub3_old)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb4_old = LBOUND(p,4)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub4_old = UBOUND(p,4)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb4 = MAX(lb4_new,lb4_old)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub4 = MIN(ub4_new,ub4_old)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4 ))
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4) = p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new,lb3_new:ub3_new,lb4_new:ub4_new))
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4) = work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 181 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_c4

! **************************************************************************************************
!> \brief (Re)Allocate an integer vector with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_i1(p, lb1_new, ub1_new)

      INTEGER, DIMENSION(:), POINTER                     :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_i1', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: zero = 0

      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: work


# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1 ))
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1) = p(lb1:ub1)
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new))
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1) = work(lb1:ub1)
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 204 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_i1

! **************************************************************************************************
!> \brief (Re)Allocate an integer array with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_i2(p, lb1_new, ub1_new, lb2_new, ub2_new)

      INTEGER, DIMENSION(:, :), POINTER                  :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_i2', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: zero = 0

      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: work


# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb2, lb2_old, ub2, ub2_old
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2_old = LBOUND(p,2)
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2_old = UBOUND(p,2)
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2 = MAX(lb2_new,lb2_old)
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2 = MIN(ub2_new,ub2_old)
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1,lb2:ub2 ))
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1,lb2:ub2) = p(lb1:ub1,lb2:ub2)
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new))
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1,lb2:ub2) = work(lb1:ub1,lb2:ub2)
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 229 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_i2

! **************************************************************************************************
!> \brief (Re)Allocate an integer array 3D with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \param lb3_new ...
!> \param ub3_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_i3(p, lb1_new, ub1_new, lb2_new, ub2_new, lb3_new, ub3_new)

      INTEGER, DIMENSION(:, :, :), POINTER               :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new, &
                                                            lb3_new, ub3_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_i3', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: zero = 0

      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: work


# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb2, lb2_old, ub2, ub2_old
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb3, lb3_old, ub3, ub3_old
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2_old = LBOUND(p,2)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2_old = UBOUND(p,2)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2 = MAX(lb2_new,lb2_old)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2 = MIN(ub2_new,ub2_old)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb3_old = LBOUND(p,3)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub3_old = UBOUND(p,3)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb3 = MAX(lb3_new,lb3_old)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub3 = MIN(ub3_new,ub3_old)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1,lb2:ub2,lb3:ub3 ))
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1,lb2:ub2,lb3:ub3) = p(lb1:ub1,lb2:ub2,lb3:ub3)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new,lb3_new:ub3_new))
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1,lb2:ub2,lb3:ub3) = work(lb1:ub1,lb2:ub2,lb3:ub3)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 257 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_i3

! **************************************************************************************************
!> \brief (Re)Allocate an integer array 3D with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \param lb3_new ...
!> \param ub3_new ...
!> \param lb4_new ...
!> \param ub4_new ...
!> \par History
!>      none
!> \author Matthias Krack (04.10.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_i4(p, lb1_new, ub1_new, lb2_new, ub2_new, lb3_new, ub3_new, &
                            lb4_new, ub4_new)

      INTEGER, DIMENSION(:, :, :, :), POINTER            :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new, &
                                                            lb3_new, ub3_new, lb4_new, ub4_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_i4', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: zero = 0

      INTEGER, ALLOCATABLE, DIMENSION(:, :, :, :)        :: work


# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb2, lb2_old, ub2, ub2_old
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb3, lb3_old, ub3, ub3_old
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb4, lb4_old, ub4, ub4_old
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2_old = LBOUND(p,2)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2_old = UBOUND(p,2)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2 = MAX(lb2_new,lb2_old)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2 = MIN(ub2_new,ub2_old)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb3_old = LBOUND(p,3)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub3_old = UBOUND(p,3)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb3 = MAX(lb3_new,lb3_old)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub3 = MIN(ub3_new,ub3_old)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb4_old = LBOUND(p,4)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub4_old = UBOUND(p,4)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb4 = MAX(lb4_new,lb4_old)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub4 = MIN(ub4_new,ub4_old)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4 ))
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4) = p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new,lb3_new:ub3_new,lb4_new:ub4_new))
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4) = work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 288 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_i4

! **************************************************************************************************
!> \brief (Re)Allocate an integer (int_8) vector with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_8i1(p, lb1_new, ub1_new)

      INTEGER(KIND=int_8), DIMENSION(:), POINTER         :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_8i1', routineP = moduleN//':'//routineN
      INTEGER(KIND=int_8), PARAMETER                     :: zero = 0

      INTEGER(KIND=int_8), ALLOCATABLE, DIMENSION(:)     :: work


# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1 ))
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1) = p(lb1:ub1)
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new))
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1) = work(lb1:ub1)
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 311 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_8i1

! **************************************************************************************************
!> \brief (Re)Allocate an integer (int_8) array with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_8i2(p, lb1_new, ub1_new, lb2_new, ub2_new)

      INTEGER(kind=int_8), DIMENSION(:, :), POINTER      :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_8i2', routineP = moduleN//':'//routineN
      INTEGER(KIND=int_8), PARAMETER                     :: zero = 0

      INTEGER(KIND=int_8), ALLOCATABLE, DIMENSION(:, :)  :: work


# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb2, lb2_old, ub2, ub2_old
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2_old = LBOUND(p,2)
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2_old = UBOUND(p,2)
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2 = MAX(lb2_new,lb2_old)
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2 = MIN(ub2_new,ub2_old)
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1,lb2:ub2 ))
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1,lb2:ub2) = p(lb1:ub1,lb2:ub2)
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new))
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1,lb2:ub2) = work(lb1:ub1,lb2:ub2)
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 336 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_8i2

! **************************************************************************************************
!> \brief (Re)Allocate an real vector with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_r1(p, lb1_new, ub1_new)

      REAL(KIND=dp), DIMENSION(:), POINTER               :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_r1', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: work


# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1 ))
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1) = p(lb1:ub1)
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new))
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1) = work(lb1:ub1)
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 359 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_r1

! **************************************************************************************************
!> \brief (Re)Allocate an real array with new dimensions.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_r2(p, lb1_new, ub1_new, lb2_new, ub2_new)

      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_r2', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: work


# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb2, lb2_old, ub2, ub2_old
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2_old = LBOUND(p,2)
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2_old = UBOUND(p,2)
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2 = MAX(lb2_new,lb2_old)
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2 = MIN(ub2_new,ub2_old)
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1,lb2:ub2 ))
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1,lb2:ub2) = p(lb1:ub1,lb2:ub2)
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new))
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1,lb2:ub2) = work(lb1:ub1,lb2:ub2)
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 384 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_r2

! **************************************************************************************************
!> \brief (Re)Allocate an real array 3D with new dimensions.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \param lb3_new ...
!> \param ub3_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_r3(p, lb1_new, ub1_new, lb2_new, ub2_new, lb3_new, ub3_new)

      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new, &
                                                            lb3_new, ub3_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_r3', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: work


# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb2, lb2_old, ub2, ub2_old
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb3, lb3_old, ub3, ub3_old
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2_old = LBOUND(p,2)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2_old = UBOUND(p,2)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2 = MAX(lb2_new,lb2_old)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2 = MIN(ub2_new,ub2_old)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb3_old = LBOUND(p,3)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub3_old = UBOUND(p,3)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb3 = MAX(lb3_new,lb3_old)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub3 = MIN(ub3_new,ub3_old)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1,lb2:ub2,lb3:ub3 ))
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1,lb2:ub2,lb3:ub3) = p(lb1:ub1,lb2:ub2,lb3:ub3)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new,lb3_new:ub3_new))
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1,lb2:ub2,lb3:ub3) = work(lb1:ub1,lb2:ub2,lb3:ub3)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 412 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_r3

! **************************************************************************************************
!> \brief (Re)Allocate an real array 4D with new dimensions.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \param lb3_new ...
!> \param ub3_new ...
!> \param lb4_new ...
!> \param ub4_new ...
!> \par History
!>      none
!> \author Matthias Krack (04.10.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_r4(p, lb1_new, ub1_new, lb2_new, ub2_new, lb3_new, ub3_new, &
                            lb4_new, ub4_new)

      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new, &
                                                            lb3_new, ub3_new, lb4_new, ub4_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_r4', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: work


# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb2, lb2_old, ub2, ub2_old
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb3, lb3_old, ub3, ub3_old
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb4, lb4_old, ub4, ub4_old
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2_old = LBOUND(p,2)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2_old = UBOUND(p,2)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2 = MAX(lb2_new,lb2_old)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2 = MIN(ub2_new,ub2_old)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb3_old = LBOUND(p,3)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub3_old = UBOUND(p,3)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb3 = MAX(lb3_new,lb3_old)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub3 = MIN(ub3_new,ub3_old)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb4_old = LBOUND(p,4)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub4_old = UBOUND(p,4)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb4 = MAX(lb4_new,lb4_old)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub4 = MIN(ub4_new,ub4_old)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4 ))
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4) = p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new,lb3_new:ub3_new,lb4_new:ub4_new))
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4) = work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 443 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_r4

! **************************************************************************************************
!> \brief (Re)Allocate an real array 5D with new dimensions.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \param lb3_new ...
!> \param ub3_new ...
!> \param lb4_new ...
!> \param ub4_new ...
!> \param lb5_new ...
!> \param ub5_new ...
!> \par History
!>      none
!> \author Matthias Krack (04.10.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_r5(p, lb1_new, ub1_new, lb2_new, ub2_new, lb3_new, ub3_new, &
                            lb4_new, ub4_new, lb5_new, ub5_new)

      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new, &
                                                            lb3_new, ub3_new, lb4_new, ub4_new, &
                                                            lb5_new, ub5_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_r5', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      REAL(KIND=dp), ALLOCATABLE, &
         DIMENSION(:, :, :, :, :)                        :: work


# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb2, lb2_old, ub2, ub2_old
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb3, lb3_old, ub3, ub3_old
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb4, lb4_old, ub4, ub4_old
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb5, lb5_old, ub5, ub5_old
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2_old = LBOUND(p,2)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2_old = UBOUND(p,2)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb2 = MAX(lb2_new,lb2_old)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub2 = MIN(ub2_new,ub2_old)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb3_old = LBOUND(p,3)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub3_old = UBOUND(p,3)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb3 = MAX(lb3_new,lb3_old)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub3 = MIN(ub3_new,ub3_old)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb4_old = LBOUND(p,4)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub4_old = UBOUND(p,4)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb4 = MAX(lb4_new,lb4_old)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub4 = MIN(ub4_new,ub4_old)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb5_old = LBOUND(p,5)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub5_old = UBOUND(p,5)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb5 = MAX(lb5_new,lb5_old)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub5 = MIN(ub5_new,ub5_old)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4,lb5:ub5 ))
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4,lb5:ub5) = p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4,lb5:ub5)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new,lb3_new:ub3_new,lb4_new:ub4_new,lb5_new:ub5_new))
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4,lb5:ub5) = work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4,lb5:ub5)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 478 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_r5

! **************************************************************************************************
!> \brief (Re)Allocate a logical vector with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_l1(p, lb1_new, ub1_new)

      LOGICAL, DIMENSION(:), POINTER                     :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_l1', routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: zero = .FALSE.

      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: work


# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1 ))
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1) = p(lb1:ub1)
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new))
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1) = work(lb1:ub1)
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 501 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_l1

! **************************************************************************************************
!> \brief (Re)Allocate a vector of string variables
!> \param p_short ...
!> \param lb_new ...
!> \param ub_new ...
!> \param p_long ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
!> \note
!>      (Maybe outdated) This routine doesnt work on SUN/Solaris!!!
!>                       It should probably not be used. (Thomas Chassaing)
! **************************************************************************************************
   SUBROUTINE reallocate_s1(p_short, lb_new, ub_new, p_long)

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), OPTIONAL, POINTER                 :: p_short
      INTEGER, INTENT(IN)                                :: lb_new, ub_new
      CHARACTER(LEN=default_path_length), DIMENSION(:), &
         OPTIONAL, POINTER                               :: p_long

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_s1', routineP = moduleN//':'//routineN

      IF (PRESENT(p_short)) THEN
         CALL reallocate_ss1(p_short, lb_new, ub_new)
      ELSE IF (PRESENT(p_long)) THEN
         CALL reallocate_ls1(p_long, lb_new, ub_new)
      ELSE
         CPABORT("At least one of the two optional arguments is required")
      END IF

   END SUBROUTINE reallocate_s1

! **************************************************************************************************
!> \brief (Re)Allocate a vector of string variables of default string length.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \par History
!>      none
!> \author Thomas Chassaing
!> \author Matthias Krack (18.07.2002,MK)
!> \note
!>      (Maybe outdated) This routine doesnt work on SUN/Solaris!!!
!>                       It should probably not be used. (Thomas Chassaing)
! **************************************************************************************************
   SUBROUTINE reallocate_ss1(p, lb1_new, ub1_new)

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_ss1', routineP = moduleN//':'//routineN
      CHARACTER(LEN=default_string_length), PARAMETER    :: zero = ""

      CHARACTER(LEN=default_string_length), &
         ALLOCATABLE, DIMENSION(:)                       :: work


# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1 ))
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1) = p(lb1:ub1)
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new))
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1) = work(lb1:ub1)
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 563 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_ss1

! **************************************************************************************************
!> \brief (Re)Allocate a vector of string variables of default path length.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
!> \note
!>      (Maybe outdated) This routine doesnt work on SUN/Solaris!!!
!>                       It should probably not be used. (Thomas Chassaing)
! **************************************************************************************************
   SUBROUTINE reallocate_ls1(p, lb1_new, ub1_new)

      CHARACTER(LEN=default_path_length), DIMENSION(:), &
         POINTER                                         :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_ls1', routineP = moduleN//':'//routineN
      CHARACTER(LEN=default_path_length), PARAMETER      :: zero = ""

      CHARACTER(LEN=default_path_length), ALLOCATABLE, &
         DIMENSION(:)                                    :: work


# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
      INTEGER :: lb1, lb1_old, ub1, ub1_old
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p)) THEN
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1_old = LBOUND(p,1)
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1_old = UBOUND(p,1)
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       lb1 = MAX(lb1_new,lb1_old)
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ub1 = MIN(ub1_new,ub1_old)
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       ALLOCATE (work( lb1:ub1 ))
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       work(lb1:ub1) = p(lb1:ub1)
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (p)
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    ALLOCATE (p(lb1_new:ub1_new))
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    p = zero
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"

# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       p(lb1:ub1) = work(lb1:ub1)
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
       DEALLOCATE (work)
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"
    END IF
# 591 "/builddir/build/BUILD/cp2k-6.1/src/common/memory_utilities.F"


   END SUBROUTINE reallocate_ls1

END MODULE memory_utilities
