# 1 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2018  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Methods to operate on n-dimensional tensor blocks.
!> \author Patrick Seewald
! **************************************************************************************************
MODULE dbcsr_tensor_block

# 1 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor.fypp" 1
# 5 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor.fypp"
# 53 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor.fypp"
# 13 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F" 2
# 14 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 15 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

   USE allocate_wrap,                   ONLY: allocate_any
   USE dbcsr_api,                       ONLY: dbcsr_type_real_4, dbcsr_type_complex_4, dbcsr_type_real_8, dbcsr_type_complex_8,&
                                              dbcsr_iterator_type, &
                                              dbcsr_iterator_blocks_left,&
                                              dbcsr_iterator_next_block, &
                                              dbcsr_iterator_start,&
                                              dbcsr_iterator_stop, &
                                              dbcsr_get_block_p, &
                                              dbcsr_put_block,&
                                              dbcsr_reserve_blocks,&
                                              dbcsr_type
   USE kinds,                           ONLY: real_4, real_8
   USE dbcsr_tensor_index,              ONLY: nd_to_2d_mapping, &
                                              ndims_mapping, &
                                              get_nd_indices, &
                                              destroy_nd_to_2d_mapping,&
                                              get_2d_indices, &
                                              get_mapping_info,&
                                              create_nd_to_2d_mapping
   USE array_list_methods,              ONLY: array_list, &
                                              get_array_elements, &
                                              destroy_array_list, &
                                              sizes_of_arrays, &
                                              create_array_list, &
                                              get_arrays
   USE dbcsr_tensor_types,              ONLY: dbcsr_t_type,&
                                              ndims_tensor,&
                                              dbcsr_t_get_data_type,&
                                              dbcsr_t_blk_sizes
#include "../base/base_uses.f90"

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_tensor_block'

   PUBLIC :: &
      block_nd, &
      create_block, &
      dbcsr_t_get_block, &
      dbcsr_t_iterator_blocks_left, &
      dbcsr_t_iterator_next_block, &
      dbcsr_t_iterator_start, &
      dbcsr_t_iterator_stop, &
      dbcsr_t_iterator_type, &
      dbcsr_t_put_block, &
      dbcsr_t_reserve_blocks,&
      destroy_block, &
      ndims_iterator,&
      reshape_2d_to_nd_block, &
      reshape_nd_to_2d_block

   TYPE dbcsr_t_iterator_type
      TYPE(dbcsr_iterator_type)   :: iter
      TYPE(nd_to_2d_mapping)      :: nd_index_blk
      TYPE(nd_to_2d_mapping)      :: nd_index
      TYPE(array_list)            :: blk_sizes, blk_offsets
   END TYPE dbcsr_t_iterator_type


# 76 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   PUBLIC :: block_nd_r_dp
# 76 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   PUBLIC :: block_nd_r_sp
# 76 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   PUBLIC :: block_nd_c_dp
# 76 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   PUBLIC :: block_nd_c_sp
# 78 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

# 80 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   TYPE block_nd_r_dp
      INTEGER, DIMENSION(:), ALLOCATABLE   :: sizes
      REAL(kind=real_8), DIMENSION(:), ALLOCATABLE :: blk
   END TYPE

# 80 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   TYPE block_nd_r_sp
      INTEGER, DIMENSION(:), ALLOCATABLE   :: sizes
      REAL(kind=real_4), DIMENSION(:), ALLOCATABLE :: blk
   END TYPE

# 80 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   TYPE block_nd_c_dp
      INTEGER, DIMENSION(:), ALLOCATABLE   :: sizes
      COMPLEX(kind=real_8), DIMENSION(:), ALLOCATABLE :: blk
   END TYPE

# 80 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   TYPE block_nd_c_sp
      INTEGER, DIMENSION(:), ALLOCATABLE   :: sizes
      COMPLEX(kind=real_4), DIMENSION(:), ALLOCATABLE :: blk
   END TYPE

# 86 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

   TYPE block_nd
# 89 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      TYPE(block_nd_r_dp) :: r_dp
# 89 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      TYPE(block_nd_r_sp) :: r_sp
# 89 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      TYPE(block_nd_c_dp) :: c_dp
# 89 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      TYPE(block_nd_c_sp) :: c_sp
# 91 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      INTEGER          :: data_type = -1
   END TYPE


   INTERFACE create_block
# 97 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE create_block_data_r_dp
# 97 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE create_block_data_r_sp
# 97 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE create_block_data_c_dp
# 97 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE create_block_data_c_sp
# 99 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE create_block_nodata
   END INTERFACE


   INTERFACE reshape_nd_to_2d_block
# 105 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 106 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_t2m_2d_r_dp
# 106 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_t2m_3d_r_dp
# 106 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_t2m_4d_r_dp
# 108 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 105 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 106 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_t2m_2d_r_sp
# 106 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_t2m_3d_r_sp
# 106 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_t2m_4d_r_sp
# 108 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 105 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 106 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_t2m_2d_c_dp
# 106 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_t2m_3d_c_dp
# 106 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_t2m_4d_c_dp
# 108 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 105 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 106 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_t2m_2d_c_sp
# 106 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_t2m_3d_c_sp
# 106 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_t2m_4d_c_sp
# 108 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 109 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   END INTERFACE reshape_nd_to_2d_block

   INTERFACE reshape_2d_to_nd_block
# 113 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 114 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_m2t_2d_r_dp
# 114 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_m2t_3d_r_dp
# 114 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_m2t_4d_r_dp
# 116 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 113 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 114 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_m2t_2d_r_sp
# 114 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_m2t_3d_r_sp
# 114 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_m2t_4d_r_sp
# 116 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 113 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 114 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_m2t_2d_c_dp
# 114 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_m2t_3d_c_dp
# 114 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_m2t_4d_c_dp
# 116 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 113 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 114 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_m2t_2d_c_sp
# 114 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_m2t_3d_c_sp
# 114 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE reshape_block_m2t_4d_c_sp
# 116 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 117 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   END INTERFACE reshape_2d_to_nd_block

   INTERFACE dbcsr_t_put_block
# 121 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 122 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_put_2d_block_r_dp
# 122 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_put_3d_block_r_dp
# 122 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_put_4d_block_r_dp
# 124 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 121 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 122 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_put_2d_block_r_sp
# 122 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_put_3d_block_r_sp
# 122 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_put_4d_block_r_sp
# 124 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 121 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 122 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_put_2d_block_c_dp
# 122 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_put_3d_block_c_dp
# 122 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_put_4d_block_c_dp
# 124 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 121 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 122 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_put_2d_block_c_sp
# 122 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_put_3d_block_c_sp
# 122 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_put_4d_block_c_sp
# 124 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 125 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_put_anyd_block
   END INTERFACE

   INTERFACE dbcsr_t_get_block
# 130 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 131 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_get_2d_block_r_dp
      MODULE PROCEDURE dbcsr_t_allocate_and_get_2d_block_r_dp
# 131 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_get_3d_block_r_dp
      MODULE PROCEDURE dbcsr_t_allocate_and_get_3d_block_r_dp
# 131 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_get_4d_block_r_dp
      MODULE PROCEDURE dbcsr_t_allocate_and_get_4d_block_r_dp
# 134 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 130 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 131 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_get_2d_block_r_sp
      MODULE PROCEDURE dbcsr_t_allocate_and_get_2d_block_r_sp
# 131 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_get_3d_block_r_sp
      MODULE PROCEDURE dbcsr_t_allocate_and_get_3d_block_r_sp
# 131 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_get_4d_block_r_sp
      MODULE PROCEDURE dbcsr_t_allocate_and_get_4d_block_r_sp
# 134 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 130 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 131 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_get_2d_block_c_dp
      MODULE PROCEDURE dbcsr_t_allocate_and_get_2d_block_c_dp
# 131 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_get_3d_block_c_dp
      MODULE PROCEDURE dbcsr_t_allocate_and_get_3d_block_c_dp
# 131 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_get_4d_block_c_dp
      MODULE PROCEDURE dbcsr_t_allocate_and_get_4d_block_c_dp
# 134 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 130 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 131 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_get_2d_block_c_sp
      MODULE PROCEDURE dbcsr_t_allocate_and_get_2d_block_c_sp
# 131 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_get_3d_block_c_sp
      MODULE PROCEDURE dbcsr_t_allocate_and_get_3d_block_c_sp
# 131 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_get_4d_block_c_sp
      MODULE PROCEDURE dbcsr_t_allocate_and_get_4d_block_c_sp
# 134 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 135 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      MODULE PROCEDURE dbcsr_t_get_anyd_block
   END INTERFACE

   INTERFACE dbcsr_t_reserve_blocks
      MODULE PROCEDURE dbcsr_t_reserve_blocks_index
      MODULE PROCEDURE dbcsr_t_reserve_blocks_index_list
      MODULE PROCEDURE dbcsr_t_reserve_blocks_index_array
      MODULE PROCEDURE dbcsr_t_reserve_blocks_template
      MODULE PROCEDURE dbcsr_t_reserve_blocks_matrix
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief Create block without data
!> \param block ...
!> \param sizes ...
!> \param data_type ...
! **************************************************************************************************
   SUBROUTINE create_block_nodata(block, sizes, data_type)
      TYPE(block_nd), INTENT(OUT)       :: block
      INTEGER, DIMENSION(:), INTENT(IN) :: sizes
      INTEGER, INTENT(IN)               :: data_type

      block%data_type = data_type
      SELECT CASE (data_type)
# 162 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_real_8)
         CALL create_block_nodata_r_dp(block%r_dp, sizes)
# 162 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_real_4)
         CALL create_block_nodata_r_sp(block%r_sp, sizes)
# 162 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_complex_8)
         CALL create_block_nodata_c_dp(block%c_dp, sizes)
# 162 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_complex_4)
         CALL create_block_nodata_c_sp(block%c_sp, sizes)
# 165 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      END SELECT
   END SUBROUTINE

! **************************************************************************************************
!> \brief Destroy block
!> \param block ...
! **************************************************************************************************
   SUBROUTINE destroy_block(block)
      TYPE(block_nd), INTENT(INOUT) :: block

      SELECT CASE (block%data_type)
# 177 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_real_8)
         CALL destroy_block_r_dp(block%r_dp)
# 177 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_real_4)
         CALL destroy_block_r_sp(block%r_sp)
# 177 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_complex_8)
         CALL destroy_block_c_dp(block%c_dp)
# 177 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_complex_4)
         CALL destroy_block_c_sp(block%c_sp)
# 180 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      END SELECT

   END SUBROUTINE

! **************************************************************************************************
!> \brief block size
!> \param block ...
!> \return ...
! **************************************************************************************************
   FUNCTION block_size(block)
      TYPE(block_nd), INTENT(IN)         :: block
      INTEGER, ALLOCATABLE, DIMENSION(:) :: block_size

      SELECT CASE (block%data_type)
# 195 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_real_8)
         CALL allocate_any(block_size, source=block%r_dp%sizes)
# 195 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_real_4)
         CALL allocate_any(block_size, source=block%r_sp%sizes)
# 195 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_complex_8)
         CALL allocate_any(block_size, source=block%c_dp%sizes)
# 195 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_complex_4)
         CALL allocate_any(block_size, source=block%c_sp%sizes)
# 198 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      END SELECT
   END FUNCTION

! **************************************************************************************************
!> \brief Generalization of dbcsr_iterator_start for tensors.
!> \param iterator ...
!> \param tensor ...
! **************************************************************************************************
   SUBROUTINE dbcsr_t_iterator_start(iterator, tensor)
      TYPE(dbcsr_t_iterator_type), INTENT(OUT)           :: iterator
      TYPE(dbcsr_t_type), INTENT(IN)                     :: tensor
      INTEGER                                            :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_iterator_start', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      CPASSERT(tensor%valid)

      CALL dbcsr_iterator_start(iterator%iter, tensor%matrix_rep)
      iterator%nd_index_blk = tensor%nd_index_blk
      iterator%nd_index = tensor%nd_index
      iterator%blk_sizes = tensor%blk_sizes
      iterator%blk_offsets = tensor%blk_offsets

      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief Generalization of dbcsr_iterator_stop for tensors.
!> \param iterator ...
! **************************************************************************************************
   SUBROUTINE dbcsr_t_iterator_stop(iterator)
      TYPE(dbcsr_t_iterator_type), INTENT(INOUT) :: iterator
      INTEGER                                    :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_iterator_stop', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      CALL dbcsr_iterator_stop(iterator%iter)
      CALL destroy_nd_to_2d_mapping(iterator%nd_index)
      CALL destroy_nd_to_2d_mapping(iterator%nd_index_blk)
      CALL destroy_array_list(iterator%blk_sizes)
      CALL destroy_array_list(iterator%blk_offsets)

      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief Number of dimensions.
!> \param iterator ...
!> \return ...
!> \par Note: specification function below must be defined before it is used in
!>      the source due to a bug in the IBM XL Fortran compiler (compilation fails)
! **************************************************************************************************
   PURE FUNCTION ndims_iterator(iterator)
      TYPE(dbcsr_t_iterator_type), INTENT(IN) :: iterator
      INTEGER                                 :: ndims_iterator

      ndims_iterator = iterator%nd_index%ndim_nd
   END FUNCTION

! **************************************************************************************************
!> \brief iterate over nd blocks of an nd rank tensor, index only (blocks must be retrieved by calling
!>        dbcsr_t_get_block on tensor).
!> \param iterator ...
!> \param ind_nd nd index of block
!> \param blk is this needed?
!> \param blk_p is this needed?
!> \param blk_size blk size in each dimension
!> \param blk_offset blk offset in each dimension
! **************************************************************************************************
   SUBROUTINE dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_p, blk_size, blk_offset)
      TYPE(dbcsr_t_iterator_type), INTENT(INOUT)     :: iterator
      INTEGER, DIMENSION(ndims_iterator(iterator)), &
         INTENT(OUT)                                 :: ind_nd
      INTEGER, INTENT(OUT)                           :: blk
      INTEGER, INTENT(OUT), OPTIONAL                 :: blk_p
      INTEGER, DIMENSION(ndims_iterator(iterator)), &
         INTENT(OUT), OPTIONAL                       :: blk_size, blk_offset

      INTEGER, DIMENSION(2)                          :: ind_2d
      INTEGER                                        :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_iterator_next_block', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      CALL dbcsr_iterator_next_block(iterator%iter, ind_2d(1), ind_2d(2), blk, blk_p=blk_p)

      ind_nd(:) = get_nd_indices(iterator%nd_index_blk, ind_2d)
      IF (PRESENT(blk_size)) blk_size(:) = get_array_elements(iterator%blk_sizes, ind_nd)
      ! note: blk_offset needs to be determined by tensor metadata, can not be derived from 2d row/col
      ! offset since block index mapping is not consistent with element index mapping
      IF (PRESENT(blk_offset)) blk_offset(:) = get_array_elements(iterator%blk_offsets, ind_nd)

      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief Generalization of dbcsr_iterator_blocks_left for tensors.
!> \param iterator ...
!> \return ...
! **************************************************************************************************
   FUNCTION dbcsr_t_iterator_blocks_left(iterator)
      TYPE(dbcsr_t_iterator_type), INTENT(IN) :: iterator
      LOGICAL                                 :: dbcsr_t_iterator_blocks_left
      INTEGER                                 :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_iterator_blocks_left', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      dbcsr_t_iterator_blocks_left = dbcsr_iterator_blocks_left(iterator%iter)

      CALL timestop(handle)
   END FUNCTION

   SUBROUTINE dbcsr_t_reserve_blocks_index_list(tensor, blk_ind)
      TYPE(dbcsr_t_type), INTENT(INOUT)  :: tensor
      TYPE(array_list), INTENT(IN)       :: blk_ind
      INTEGER, DIMENSION(:), ALLOCATABLE :: blk_ind_1, blk_ind_2, blk_ind_3, blk_ind_4
      INTEGER                            :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_reserve_blocks_index_list', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
# 323 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      IF (ndims_tensor(tensor) == 2) THEN
         CALL get_arrays(blk_ind, blk_ind_1, blk_ind_2)
         CALL dbcsr_t_reserve_blocks(tensor, blk_ind_1, blk_ind_2)
      ENDIF
# 323 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      IF (ndims_tensor(tensor) == 3) THEN
         CALL get_arrays(blk_ind, blk_ind_1, blk_ind_2, blk_ind_3)
         CALL dbcsr_t_reserve_blocks(tensor, blk_ind_1, blk_ind_2, blk_ind_3)
      ENDIF
# 323 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      IF (ndims_tensor(tensor) == 4) THEN
         CALL get_arrays(blk_ind, blk_ind_1, blk_ind_2, blk_ind_3, blk_ind_4)
         CALL dbcsr_t_reserve_blocks(tensor, blk_ind_1, blk_ind_2, blk_ind_3, blk_ind_4)
      ENDIF
# 328 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CALL timestop(handle)
   END SUBROUTINE

   SUBROUTINE dbcsr_t_reserve_blocks_index_array(tensor, blk_ind)
      TYPE(dbcsr_t_type), INTENT(INOUT)   :: tensor
      INTEGER, DIMENSION(:,:), INTENT(IN) :: blk_ind
      INTEGER                             :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_reserve_blocks_index_array', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
# 340 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      IF (ndims_tensor(tensor) == 2) THEN
         CALL dbcsr_t_reserve_blocks(tensor, blk_ind(:,1), blk_ind(:,2))
      ENDIF
# 340 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      IF (ndims_tensor(tensor) == 3) THEN
         CALL dbcsr_t_reserve_blocks(tensor, blk_ind(:,1), blk_ind(:,2), blk_ind(:,3))
      ENDIF
# 340 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      IF (ndims_tensor(tensor) == 4) THEN
         CALL dbcsr_t_reserve_blocks(tensor, blk_ind(:,1), blk_ind(:,2), blk_ind(:,3), blk_ind(:,4))
      ENDIF
# 344 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief reserve tensor blocks using block indices
!> \param tensor ...
!> \param blk_ind_1 index of blocks to reserve in each dimension
!> \param blk_ind_2 ...
!> \param blk_ind_3 ...
! **************************************************************************************************
   SUBROUTINE dbcsr_t_reserve_blocks_index(tensor, blk_ind_1, blk_ind_2, blk_ind_3, blk_ind_4)
      TYPE(dbcsr_t_type), INTENT(INOUT)           :: tensor
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL :: blk_ind_1, blk_ind_2, blk_ind_3, blk_ind_4
      INTEGER                                     :: iblk, nblk, handle
      INTEGER, ALLOCATABLE, DIMENSION(:)          :: cols, rows
      INTEGER, DIMENSION(2)                       :: ind_2d
      TYPE(array_list)                            :: blks
      INTEGER, DIMENSION(ndims_tensor(tensor))   :: iblk_nd, ind_nd, nblk_tmp
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_reserve_blocks_index', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      CPASSERT(tensor%valid)

      CALL create_array_list(blks, ndims_tensor(tensor), &
                             blk_ind_1, blk_ind_2, blk_ind_3, blk_ind_4)

      nblk_tmp(:) = sizes_of_arrays(blks)
      nblk = nblk_tmp(1)
      ALLOCATE (cols(nblk), rows(nblk))
      DO iblk = 1, nblk
         iblk_nd(:) = iblk
         ind_nd(:) = get_array_elements(blks, iblk_nd)
         ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind_nd)
         rows(iblk) = ind_2d(1); cols(iblk) = ind_2d(2)
      ENDDO

      CALL dbcsr_reserve_blocks(tensor%matrix_rep, rows=rows, cols=cols)
      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief reserve tensor blocks using template
!> \param tensor_in template tensor
!> \param tensor_out ...
! **************************************************************************************************
   SUBROUTINE dbcsr_t_reserve_blocks_template(tensor_in, tensor_out)
      TYPE(dbcsr_t_type), INTENT(IN)    :: tensor_in
      TYPE(dbcsr_t_type), INTENT(INOUT) :: tensor_out
      INTEGER                           :: handle

      TYPE(array_list)                  :: blk_ind
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_reserve_blocks_template', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)

      CALL dbcsr_t_reserved_block_indices(tensor_in, blk_ind)
      CALL dbcsr_t_reserve_blocks(tensor_out, blk_ind)

      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief reserve tensor blocks using matrix template
!> \param matrix_in ...
!> \param tensor_out ...
! **************************************************************************************************
   SUBROUTINE dbcsr_t_reserve_blocks_matrix(matrix_in, tensor_out)
      TYPE(dbcsr_type), INTENT(IN)        :: matrix_in
      TYPE(dbcsr_t_type), INTENT(INOUT)  :: tensor_out

      INTEGER                            :: blk, iblk, nblk
      INTEGER, ALLOCATABLE, DIMENSION(:) :: blk_ind_1, blk_ind_2
      INTEGER, DIMENSION(2)              :: ind_2d
      TYPE(dbcsr_iterator_type)          :: iter
      INTEGER                            :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_reserve_blocks_matrix', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      nblk = 0
      CALL dbcsr_iterator_start(iter, matrix_in)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, ind_2d(1), ind_2d(2), blk)
         nblk = nblk + 1
      ENDDO
      CALL dbcsr_iterator_stop(iter)

      ALLOCATE (blk_ind_1(nblk), blk_ind_2(nblk))
      CALL dbcsr_iterator_start(iter, matrix_in)
      DO iblk = 1, nblk
         CALL dbcsr_iterator_next_block(iter, ind_2d(1), ind_2d(2), blk)
         blk_ind_1(iblk) = ind_2d(1); blk_ind_2(iblk) = ind_2d(2)
      ENDDO
      CALL dbcsr_iterator_stop(iter)

      CALL dbcsr_t_reserve_blocks(tensor_out, blk_ind_1, blk_ind_2)
      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief Number of non-zero blocks
!> \param tensor ...
!> \return ...
! **************************************************************************************************
   FUNCTION nblocks_reserved(tensor)
      TYPE(dbcsr_t_type), INTENT(IN)            :: tensor
      INTEGER                                   :: nblocks_reserved

      INTEGER                                   :: blk, blk_p
      TYPE(dbcsr_t_iterator_type)               :: iterator
      INTEGER, DIMENSION(ndims_tensor(tensor)) :: blk_offset, blk_size, ind_nd

      CPASSERT(tensor%valid)

      CALL dbcsr_t_iterator_start(iterator, tensor)
      nblocks_reserved = 0
      DO WHILE (dbcsr_t_iterator_blocks_left(iterator))
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk, blk_p, blk_size, blk_offset)
         nblocks_reserved = nblocks_reserved + 1
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)
   END FUNCTION

! **************************************************************************************************
!> \brief indices of non-zero blocks
!> \param tensor ...
!> \param blk_ind_1 ...
!> \param blk_ind_2 ...
!> \param blk_ind_3 ...
! **************************************************************************************************
   SUBROUTINE dbcsr_t_reserved_block_indices(tensor, blk_ind)
      TYPE(dbcsr_t_type), INTENT(IN)            :: tensor
      INTEGER, ALLOCATABLE, DIMENSION(:)        :: blk_ind_1, blk_ind_2, blk_ind_3, blk_ind_4
      INTEGER                                   :: blk, iblk, nblk
      TYPE(dbcsr_t_iterator_type)               :: iterator
      INTEGER, DIMENSION(ndims_tensor(tensor))  :: ind_nd
      TYPE(array_list), INTENT(OUT)             :: blk_ind

      CPASSERT(tensor%valid)

      nblk = nblocks_reserved(tensor)

# 490 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      IF (ndims_tensor(tensor) >= 1) ALLOCATE (blk_ind_1(nblk))
# 490 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      IF (ndims_tensor(tensor) >= 2) ALLOCATE (blk_ind_2(nblk))
# 490 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      IF (ndims_tensor(tensor) >= 3) ALLOCATE (blk_ind_3(nblk))
# 490 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      IF (ndims_tensor(tensor) >= 4) ALLOCATE (blk_ind_4(nblk))
# 492 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

      CALL dbcsr_t_iterator_start(iterator, tensor)
      DO iblk = 1, nblk
         CALL dbcsr_t_iterator_next_block(iterator, ind_nd, blk)
# 497 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
         IF (ndims_tensor(tensor) >= 1) blk_ind_1(iblk) = ind_nd(1)
# 497 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
         IF (ndims_tensor(tensor) >= 2) blk_ind_2(iblk) = ind_nd(2)
# 497 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
         IF (ndims_tensor(tensor) >= 3) blk_ind_3(iblk) = ind_nd(3)
# 497 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
         IF (ndims_tensor(tensor) >= 4) blk_ind_4(iblk) = ind_nd(4)
# 499 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      ENDDO
      CALL dbcsr_t_iterator_stop(iterator)

# 503 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      IF (ndims_tensor(tensor) == 2) THEN
         CALL create_array_list(blk_ind, ndims_tensor(tensor), blk_ind_1, blk_ind_2)
      ENDIF
# 503 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      IF (ndims_tensor(tensor) == 3) THEN
         CALL create_array_list(blk_ind, ndims_tensor(tensor), blk_ind_1, blk_ind_2, blk_ind_3)
      ENDIF
# 503 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      IF (ndims_tensor(tensor) == 4) THEN
         CALL create_array_list(blk_ind, ndims_tensor(tensor), blk_ind_1, blk_ind_2, blk_ind_3, blk_ind_4)
      ENDIF
# 507 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

   END SUBROUTINE

# 511 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Create block from array, array can be n-dimensional.
!> \param block ...
!> \param sizes ...
!> \param array ...
! **************************************************************************************************
   SUBROUTINE create_block_data_r_dp(block, sizes, array)
      TYPE(block_nd), INTENT(OUT)                       :: block
      INTEGER, DIMENSION(:), INTENT(IN)                 :: sizes
      REAL(kind=real_8), DIMENSION(PRODUCT(sizes)),  INTENT(IN) :: array

      ASSOCIATE(blk => block%r_dp)
      block%data_type = dbcsr_type_real_8
      CALL allocate_any(blk%sizes, source=sizes)
      CALL allocate_any(blk%blk, source=array)
      END ASSOCIATE
   END SUBROUTINE
# 511 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Create block from array, array can be n-dimensional.
!> \param block ...
!> \param sizes ...
!> \param array ...
! **************************************************************************************************
   SUBROUTINE create_block_data_r_sp(block, sizes, array)
      TYPE(block_nd), INTENT(OUT)                       :: block
      INTEGER, DIMENSION(:), INTENT(IN)                 :: sizes
      REAL(kind=real_4), DIMENSION(PRODUCT(sizes)),  INTENT(IN) :: array

      ASSOCIATE(blk => block%r_sp)
      block%data_type = dbcsr_type_real_4
      CALL allocate_any(blk%sizes, source=sizes)
      CALL allocate_any(blk%blk, source=array)
      END ASSOCIATE
   END SUBROUTINE
# 511 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Create block from array, array can be n-dimensional.
!> \param block ...
!> \param sizes ...
!> \param array ...
! **************************************************************************************************
   SUBROUTINE create_block_data_c_dp(block, sizes, array)
      TYPE(block_nd), INTENT(OUT)                       :: block
      INTEGER, DIMENSION(:), INTENT(IN)                 :: sizes
      COMPLEX(kind=real_8), DIMENSION(PRODUCT(sizes)),  INTENT(IN) :: array

      ASSOCIATE(blk => block%c_dp)
      block%data_type = dbcsr_type_complex_8
      CALL allocate_any(blk%sizes, source=sizes)
      CALL allocate_any(blk%blk, source=array)
      END ASSOCIATE
   END SUBROUTINE
# 511 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Create block from array, array can be n-dimensional.
!> \param block ...
!> \param sizes ...
!> \param array ...
! **************************************************************************************************
   SUBROUTINE create_block_data_c_sp(block, sizes, array)
      TYPE(block_nd), INTENT(OUT)                       :: block
      INTEGER, DIMENSION(:), INTENT(IN)                 :: sizes
      COMPLEX(kind=real_4), DIMENSION(PRODUCT(sizes)),  INTENT(IN) :: array

      ASSOCIATE(blk => block%c_sp)
      block%data_type = dbcsr_type_complex_4
      CALL allocate_any(blk%sizes, source=sizes)
      CALL allocate_any(blk%blk, source=array)
      END ASSOCIATE
   END SUBROUTINE
# 529 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

# 531 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Create and allocate block, but no data.
!> \param block ...
!> \param sizes ...
! **************************************************************************************************
   SUBROUTINE create_block_nodata_r_dp(block, sizes)
      INTEGER, INTENT(IN), DIMENSION(:)       :: sizes
      TYPE(block_nd_r_dp), INTENT(OUT) :: block
      CALL allocate_any(block%sizes, source=sizes)
      ALLOCATE(block%blk(PRODUCT(sizes)))
   END SUBROUTINE
# 531 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Create and allocate block, but no data.
!> \param block ...
!> \param sizes ...
! **************************************************************************************************
   SUBROUTINE create_block_nodata_r_sp(block, sizes)
      INTEGER, INTENT(IN), DIMENSION(:)       :: sizes
      TYPE(block_nd_r_sp), INTENT(OUT) :: block
      CALL allocate_any(block%sizes, source=sizes)
      ALLOCATE(block%blk(PRODUCT(sizes)))
   END SUBROUTINE
# 531 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Create and allocate block, but no data.
!> \param block ...
!> \param sizes ...
! **************************************************************************************************
   SUBROUTINE create_block_nodata_c_dp(block, sizes)
      INTEGER, INTENT(IN), DIMENSION(:)       :: sizes
      TYPE(block_nd_c_dp), INTENT(OUT) :: block
      CALL allocate_any(block%sizes, source=sizes)
      ALLOCATE(block%blk(PRODUCT(sizes)))
   END SUBROUTINE
# 531 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Create and allocate block, but no data.
!> \param block ...
!> \param sizes ...
! **************************************************************************************************
   SUBROUTINE create_block_nodata_c_sp(block, sizes)
      INTEGER, INTENT(IN), DIMENSION(:)       :: sizes
      TYPE(block_nd_c_sp), INTENT(OUT) :: block
      CALL allocate_any(block%sizes, source=sizes)
      ALLOCATE(block%blk(PRODUCT(sizes)))
   END SUBROUTINE
# 543 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

# 545 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE destroy_block_r_dp(block)
      TYPE(block_nd_r_dp), INTENT(INOUT) :: block
      DEALLOCATE(block%blk)
      DEALLOCATE(block%sizes)
   END SUBROUTINE
# 545 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE destroy_block_r_sp(block)
      TYPE(block_nd_r_sp), INTENT(INOUT) :: block
      DEALLOCATE(block%blk)
      DEALLOCATE(block%sizes)
   END SUBROUTINE
# 545 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE destroy_block_c_dp(block)
      TYPE(block_nd_c_dp), INTENT(INOUT) :: block
      DEALLOCATE(block%blk)
      DEALLOCATE(block%sizes)
   END SUBROUTINE
# 545 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE destroy_block_c_sp(block)
      TYPE(block_nd_c_sp), INTENT(INOUT) :: block
      DEALLOCATE(block%blk)
      DEALLOCATE(block%sizes)
   END SUBROUTINE
# 555 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

! **************************************************************************************************
!> \brief Generic implementation of dbcsr_t_get_block (arbitrary tensor rank and arbitrary datatype)
!> \param tensor ...
!> \param ind block index
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_anyd_block(tensor, ind, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)            :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), &
         INTENT(IN)                                :: ind
      TYPE(block_nd), INTENT(OUT)                  :: block
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_get_anyd_block', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      SELECT CASE (dbcsr_t_get_data_type(tensor))
# 574 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_real_8)
         CALL dbcsr_t_get_anyd_block_r_dp(tensor, ind, block)
# 574 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_real_4)
         CALL dbcsr_t_get_anyd_block_r_sp(tensor, ind, block)
# 574 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_complex_8)
         CALL dbcsr_t_get_anyd_block_c_dp(tensor, ind, block)
# 574 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_complex_4)
         CALL dbcsr_t_get_anyd_block_c_sp(tensor, ind, block)
# 577 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      END SELECT
      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief Generic implementation of dbcsr_t_put_block (arbitrary tensor rank and arbitrary datatype)
!> \param tensor ...
!> \param ind block index
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_anyd_block(tensor, ind, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)            :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), &
         INTENT(IN)                                :: ind
      TYPE(block_nd), INTENT(IN)                   :: block
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_put_anyd_block', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      SELECT CASE (block%data_type)
# 599 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_real_8)
         CALL dbcsr_t_put_anyd_block_r_dp(tensor, ind, block%r_dp)
# 599 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_real_4)
         CALL dbcsr_t_put_anyd_block_r_sp(tensor, ind, block%r_sp)
# 599 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_complex_8)
         CALL dbcsr_t_put_anyd_block_c_dp(tensor, ind, block%c_dp)
# 599 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE (dbcsr_type_complex_4)
         CALL dbcsr_t_put_anyd_block_c_sp(tensor, ind, block%c_sp)
# 602 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      END SELECT
      CALL timestop(handle)
   END SUBROUTINE

# 607 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Generic implementation of dbcsr_t_put_block, template for datatype
!> \param tensor ...
!> \param ind block index
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_anyd_block_r_dp(tensor, ind, block)
      TYPE(block_nd_r_dp), INTENT(IN)       :: block
      TYPE(dbcsr_t_type), INTENT(INOUT)            :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), &
         INTENT(IN)                                :: ind

      SELECT CASE(ndims_tensor(tensor))
# 621 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(2)
         CALL dbcsr_t_put_2d_block_r_dp(tensor, ind, block%sizes, block%blk)
# 621 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(3)
         CALL dbcsr_t_put_3d_block_r_dp(tensor, ind, block%sizes, block%blk)
# 621 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(4)
         CALL dbcsr_t_put_4d_block_r_dp(tensor, ind, block%sizes, block%blk)
# 624 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      END SELECT
   END SUBROUTINE
# 607 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Generic implementation of dbcsr_t_put_block, template for datatype
!> \param tensor ...
!> \param ind block index
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_anyd_block_r_sp(tensor, ind, block)
      TYPE(block_nd_r_sp), INTENT(IN)       :: block
      TYPE(dbcsr_t_type), INTENT(INOUT)            :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), &
         INTENT(IN)                                :: ind

      SELECT CASE(ndims_tensor(tensor))
# 621 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(2)
         CALL dbcsr_t_put_2d_block_r_sp(tensor, ind, block%sizes, block%blk)
# 621 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(3)
         CALL dbcsr_t_put_3d_block_r_sp(tensor, ind, block%sizes, block%blk)
# 621 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(4)
         CALL dbcsr_t_put_4d_block_r_sp(tensor, ind, block%sizes, block%blk)
# 624 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      END SELECT
   END SUBROUTINE
# 607 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Generic implementation of dbcsr_t_put_block, template for datatype
!> \param tensor ...
!> \param ind block index
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_anyd_block_c_dp(tensor, ind, block)
      TYPE(block_nd_c_dp), INTENT(IN)       :: block
      TYPE(dbcsr_t_type), INTENT(INOUT)            :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), &
         INTENT(IN)                                :: ind

      SELECT CASE(ndims_tensor(tensor))
# 621 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(2)
         CALL dbcsr_t_put_2d_block_c_dp(tensor, ind, block%sizes, block%blk)
# 621 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(3)
         CALL dbcsr_t_put_3d_block_c_dp(tensor, ind, block%sizes, block%blk)
# 621 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(4)
         CALL dbcsr_t_put_4d_block_c_dp(tensor, ind, block%sizes, block%blk)
# 624 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      END SELECT
   END SUBROUTINE
# 607 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Generic implementation of dbcsr_t_put_block, template for datatype
!> \param tensor ...
!> \param ind block index
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_anyd_block_c_sp(tensor, ind, block)
      TYPE(block_nd_c_sp), INTENT(IN)       :: block
      TYPE(dbcsr_t_type), INTENT(INOUT)            :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), &
         INTENT(IN)                                :: ind

      SELECT CASE(ndims_tensor(tensor))
# 621 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(2)
         CALL dbcsr_t_put_2d_block_c_sp(tensor, ind, block%sizes, block%blk)
# 621 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(3)
         CALL dbcsr_t_put_3d_block_c_sp(tensor, ind, block%sizes, block%blk)
# 621 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(4)
         CALL dbcsr_t_put_4d_block_c_sp(tensor, ind, block%sizes, block%blk)
# 624 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      END SELECT
   END SUBROUTINE
# 627 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

# 629 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Generic implementation of dbcsr_t_get_block (arbitrary tensor rank)
!> \param tensor ...
!> \param ind block index
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_anyd_block_r_dp(tensor, ind, block)
      TYPE(block_nd), INTENT(OUT)                  :: block
      TYPE(dbcsr_t_type), INTENT(INOUT)            :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), &
         INTENT(IN)                                :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor))    :: blk_size
      REAL(kind=real_8), DIMENSION(:), ALLOCATABLE         :: block_arr

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      ALLOCATE(block_arr(PRODUCT(blk_size)))

      SELECT CASE(ndims_tensor(tensor))
# 648 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(2)
         CALL dbcsr_t_get_2d_block_r_dp(tensor, ind, blk_size, block_arr)
# 648 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(3)
         CALL dbcsr_t_get_3d_block_r_dp(tensor, ind, blk_size, block_arr)
# 648 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(4)
         CALL dbcsr_t_get_4d_block_r_dp(tensor, ind, blk_size, block_arr)
# 651 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      END SELECT
      CALL create_block(block, blk_size, block_arr)
   END SUBROUTINE
# 629 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Generic implementation of dbcsr_t_get_block (arbitrary tensor rank)
!> \param tensor ...
!> \param ind block index
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_anyd_block_r_sp(tensor, ind, block)
      TYPE(block_nd), INTENT(OUT)                  :: block
      TYPE(dbcsr_t_type), INTENT(INOUT)            :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), &
         INTENT(IN)                                :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor))    :: blk_size
      REAL(kind=real_4), DIMENSION(:), ALLOCATABLE         :: block_arr

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      ALLOCATE(block_arr(PRODUCT(blk_size)))

      SELECT CASE(ndims_tensor(tensor))
# 648 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(2)
         CALL dbcsr_t_get_2d_block_r_sp(tensor, ind, blk_size, block_arr)
# 648 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(3)
         CALL dbcsr_t_get_3d_block_r_sp(tensor, ind, blk_size, block_arr)
# 648 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(4)
         CALL dbcsr_t_get_4d_block_r_sp(tensor, ind, blk_size, block_arr)
# 651 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      END SELECT
      CALL create_block(block, blk_size, block_arr)
   END SUBROUTINE
# 629 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Generic implementation of dbcsr_t_get_block (arbitrary tensor rank)
!> \param tensor ...
!> \param ind block index
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_anyd_block_c_dp(tensor, ind, block)
      TYPE(block_nd), INTENT(OUT)                  :: block
      TYPE(dbcsr_t_type), INTENT(INOUT)            :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), &
         INTENT(IN)                                :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor))    :: blk_size
      COMPLEX(kind=real_8), DIMENSION(:), ALLOCATABLE         :: block_arr

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      ALLOCATE(block_arr(PRODUCT(blk_size)))

      SELECT CASE(ndims_tensor(tensor))
# 648 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(2)
         CALL dbcsr_t_get_2d_block_c_dp(tensor, ind, blk_size, block_arr)
# 648 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(3)
         CALL dbcsr_t_get_3d_block_c_dp(tensor, ind, blk_size, block_arr)
# 648 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(4)
         CALL dbcsr_t_get_4d_block_c_dp(tensor, ind, blk_size, block_arr)
# 651 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      END SELECT
      CALL create_block(block, blk_size, block_arr)
   END SUBROUTINE
# 629 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Generic implementation of dbcsr_t_get_block (arbitrary tensor rank)
!> \param tensor ...
!> \param ind block index
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_anyd_block_c_sp(tensor, ind, block)
      TYPE(block_nd), INTENT(OUT)                  :: block
      TYPE(dbcsr_t_type), INTENT(INOUT)            :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), &
         INTENT(IN)                                :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor))    :: blk_size
      COMPLEX(kind=real_4), DIMENSION(:), ALLOCATABLE         :: block_arr

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      ALLOCATE(block_arr(PRODUCT(blk_size)))

      SELECT CASE(ndims_tensor(tensor))
# 648 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(2)
         CALL dbcsr_t_get_2d_block_c_sp(tensor, ind, blk_size, block_arr)
# 648 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(3)
         CALL dbcsr_t_get_3d_block_c_sp(tensor, ind, blk_size, block_arr)
# 648 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      CASE(4)
         CALL dbcsr_t_get_4d_block_c_sp(tensor, ind, blk_size, block_arr)
# 651 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
      END SELECT
      CALL create_block(block, blk_size, block_arr)
   END SUBROUTINE
# 655 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

# 657 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 658 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_put_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_2d_block_r_dp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      REAL(kind=real_8), DIMENSION(sizes(1), sizes(2)), &
         INTENT(IN)                                         :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d, dims_2d
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:, :)               :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_put_2d_block_r_dp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      ! reshape block
      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL get_mapping_info(map_blk, dims_2d=dims_2d)
      CALL allocate_any(block_2d, shape_spec=dims_2d)
      CALL reshape_nd_to_2d_block(map_blk, block, block_2d)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)

      CALL dbcsr_put_block(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d)

      CALL timestop(handle)
   END SUBROUTINE
# 658 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_put_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_3d_block_r_dp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      REAL(kind=real_8), DIMENSION(sizes(1), sizes(2), sizes(3)), &
         INTENT(IN)                                         :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d, dims_2d
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:, :)               :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_put_3d_block_r_dp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      ! reshape block
      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL get_mapping_info(map_blk, dims_2d=dims_2d)
      CALL allocate_any(block_2d, shape_spec=dims_2d)
      CALL reshape_nd_to_2d_block(map_blk, block, block_2d)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)

      CALL dbcsr_put_block(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d)

      CALL timestop(handle)
   END SUBROUTINE
# 658 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_put_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_4d_block_r_dp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      REAL(kind=real_8), DIMENSION(sizes(1), sizes(2), sizes(3), sizes(4)), &
         INTENT(IN)                                         :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d, dims_2d
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:, :)               :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_put_4d_block_r_dp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      ! reshape block
      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL get_mapping_info(map_blk, dims_2d=dims_2d)
      CALL allocate_any(block_2d, shape_spec=dims_2d)
      CALL reshape_nd_to_2d_block(map_blk, block, block_2d)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)

      CALL dbcsr_put_block(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d)

      CALL timestop(handle)
   END SUBROUTINE
# 696 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 657 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 658 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_put_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_2d_block_r_sp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      REAL(kind=real_4), DIMENSION(sizes(1), sizes(2)), &
         INTENT(IN)                                         :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d, dims_2d
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:, :)               :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_put_2d_block_r_sp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      ! reshape block
      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL get_mapping_info(map_blk, dims_2d=dims_2d)
      CALL allocate_any(block_2d, shape_spec=dims_2d)
      CALL reshape_nd_to_2d_block(map_blk, block, block_2d)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)

      CALL dbcsr_put_block(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d)

      CALL timestop(handle)
   END SUBROUTINE
# 658 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_put_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_3d_block_r_sp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      REAL(kind=real_4), DIMENSION(sizes(1), sizes(2), sizes(3)), &
         INTENT(IN)                                         :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d, dims_2d
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:, :)               :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_put_3d_block_r_sp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      ! reshape block
      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL get_mapping_info(map_blk, dims_2d=dims_2d)
      CALL allocate_any(block_2d, shape_spec=dims_2d)
      CALL reshape_nd_to_2d_block(map_blk, block, block_2d)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)

      CALL dbcsr_put_block(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d)

      CALL timestop(handle)
   END SUBROUTINE
# 658 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_put_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_4d_block_r_sp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      REAL(kind=real_4), DIMENSION(sizes(1), sizes(2), sizes(3), sizes(4)), &
         INTENT(IN)                                         :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d, dims_2d
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:, :)               :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_put_4d_block_r_sp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      ! reshape block
      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL get_mapping_info(map_blk, dims_2d=dims_2d)
      CALL allocate_any(block_2d, shape_spec=dims_2d)
      CALL reshape_nd_to_2d_block(map_blk, block, block_2d)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)

      CALL dbcsr_put_block(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d)

      CALL timestop(handle)
   END SUBROUTINE
# 696 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 657 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 658 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_put_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_2d_block_c_dp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      COMPLEX(kind=real_8), DIMENSION(sizes(1), sizes(2)), &
         INTENT(IN)                                         :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d, dims_2d
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:, :)               :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_put_2d_block_c_dp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      ! reshape block
      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL get_mapping_info(map_blk, dims_2d=dims_2d)
      CALL allocate_any(block_2d, shape_spec=dims_2d)
      CALL reshape_nd_to_2d_block(map_blk, block, block_2d)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)

      CALL dbcsr_put_block(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d)

      CALL timestop(handle)
   END SUBROUTINE
# 658 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_put_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_3d_block_c_dp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      COMPLEX(kind=real_8), DIMENSION(sizes(1), sizes(2), sizes(3)), &
         INTENT(IN)                                         :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d, dims_2d
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:, :)               :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_put_3d_block_c_dp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      ! reshape block
      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL get_mapping_info(map_blk, dims_2d=dims_2d)
      CALL allocate_any(block_2d, shape_spec=dims_2d)
      CALL reshape_nd_to_2d_block(map_blk, block, block_2d)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)

      CALL dbcsr_put_block(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d)

      CALL timestop(handle)
   END SUBROUTINE
# 658 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_put_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_4d_block_c_dp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      COMPLEX(kind=real_8), DIMENSION(sizes(1), sizes(2), sizes(3), sizes(4)), &
         INTENT(IN)                                         :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d, dims_2d
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:, :)               :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_put_4d_block_c_dp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      ! reshape block
      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL get_mapping_info(map_blk, dims_2d=dims_2d)
      CALL allocate_any(block_2d, shape_spec=dims_2d)
      CALL reshape_nd_to_2d_block(map_blk, block, block_2d)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)

      CALL dbcsr_put_block(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d)

      CALL timestop(handle)
   END SUBROUTINE
# 696 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 657 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 658 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_put_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_2d_block_c_sp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      COMPLEX(kind=real_4), DIMENSION(sizes(1), sizes(2)), &
         INTENT(IN)                                         :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d, dims_2d
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:, :)               :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_put_2d_block_c_sp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      ! reshape block
      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL get_mapping_info(map_blk, dims_2d=dims_2d)
      CALL allocate_any(block_2d, shape_spec=dims_2d)
      CALL reshape_nd_to_2d_block(map_blk, block, block_2d)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)

      CALL dbcsr_put_block(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d)

      CALL timestop(handle)
   END SUBROUTINE
# 658 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_put_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_3d_block_c_sp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      COMPLEX(kind=real_4), DIMENSION(sizes(1), sizes(2), sizes(3)), &
         INTENT(IN)                                         :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d, dims_2d
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:, :)               :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_put_3d_block_c_sp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      ! reshape block
      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL get_mapping_info(map_blk, dims_2d=dims_2d)
      CALL allocate_any(block_2d, shape_spec=dims_2d)
      CALL reshape_nd_to_2d_block(map_blk, block, block_2d)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)

      CALL dbcsr_put_block(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d)

      CALL timestop(handle)
   END SUBROUTINE
# 658 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_put_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to put
! **************************************************************************************************
   SUBROUTINE dbcsr_t_put_4d_block_c_sp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      COMPLEX(kind=real_4), DIMENSION(sizes(1), sizes(2), sizes(3), sizes(4)), &
         INTENT(IN)                                         :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d, dims_2d
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:, :)               :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_put_4d_block_c_sp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      ! reshape block
      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL get_mapping_info(map_blk, dims_2d=dims_2d)
      CALL allocate_any(block_2d, shape_spec=dims_2d)
      CALL reshape_nd_to_2d_block(map_blk, block, block_2d)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)

      CALL dbcsr_put_block(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d)

      CALL timestop(handle)
   END SUBROUTINE
# 696 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 697 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

# 699 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 700 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   SUBROUTINE dbcsr_t_allocate_and_get_2d_block_r_dp(tensor, ind, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN)  :: ind
      REAL(kind=real_8), DIMENSION(:,:), &
         ALLOCATABLE, INTENT(OUT)                           :: block
      INTEGER, DIMENSION(2)                          :: blk_size

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      CALL allocate_any(block, shape_spec=blk_size)
      CALL dbcsr_t_get_2d_block_r_dp(tensor, ind, blk_size, block)

   END SUBROUTINE
# 700 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   SUBROUTINE dbcsr_t_allocate_and_get_3d_block_r_dp(tensor, ind, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN)  :: ind
      REAL(kind=real_8), DIMENSION(:,:,:), &
         ALLOCATABLE, INTENT(OUT)                           :: block
      INTEGER, DIMENSION(3)                          :: blk_size

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      CALL allocate_any(block, shape_spec=blk_size)
      CALL dbcsr_t_get_3d_block_r_dp(tensor, ind, blk_size, block)

   END SUBROUTINE
# 700 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   SUBROUTINE dbcsr_t_allocate_and_get_4d_block_r_dp(tensor, ind, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN)  :: ind
      REAL(kind=real_8), DIMENSION(:,:,:,:), &
         ALLOCATABLE, INTENT(OUT)                           :: block
      INTEGER, DIMENSION(4)                          :: blk_size

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      CALL allocate_any(block, shape_spec=blk_size)
      CALL dbcsr_t_get_4d_block_r_dp(tensor, ind, blk_size, block)

   END SUBROUTINE
# 713 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 699 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 700 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   SUBROUTINE dbcsr_t_allocate_and_get_2d_block_r_sp(tensor, ind, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN)  :: ind
      REAL(kind=real_4), DIMENSION(:,:), &
         ALLOCATABLE, INTENT(OUT)                           :: block
      INTEGER, DIMENSION(2)                          :: blk_size

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      CALL allocate_any(block, shape_spec=blk_size)
      CALL dbcsr_t_get_2d_block_r_sp(tensor, ind, blk_size, block)

   END SUBROUTINE
# 700 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   SUBROUTINE dbcsr_t_allocate_and_get_3d_block_r_sp(tensor, ind, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN)  :: ind
      REAL(kind=real_4), DIMENSION(:,:,:), &
         ALLOCATABLE, INTENT(OUT)                           :: block
      INTEGER, DIMENSION(3)                          :: blk_size

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      CALL allocate_any(block, shape_spec=blk_size)
      CALL dbcsr_t_get_3d_block_r_sp(tensor, ind, blk_size, block)

   END SUBROUTINE
# 700 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   SUBROUTINE dbcsr_t_allocate_and_get_4d_block_r_sp(tensor, ind, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN)  :: ind
      REAL(kind=real_4), DIMENSION(:,:,:,:), &
         ALLOCATABLE, INTENT(OUT)                           :: block
      INTEGER, DIMENSION(4)                          :: blk_size

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      CALL allocate_any(block, shape_spec=blk_size)
      CALL dbcsr_t_get_4d_block_r_sp(tensor, ind, blk_size, block)

   END SUBROUTINE
# 713 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 699 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 700 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   SUBROUTINE dbcsr_t_allocate_and_get_2d_block_c_dp(tensor, ind, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN)  :: ind
      COMPLEX(kind=real_8), DIMENSION(:,:), &
         ALLOCATABLE, INTENT(OUT)                           :: block
      INTEGER, DIMENSION(2)                          :: blk_size

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      CALL allocate_any(block, shape_spec=blk_size)
      CALL dbcsr_t_get_2d_block_c_dp(tensor, ind, blk_size, block)

   END SUBROUTINE
# 700 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   SUBROUTINE dbcsr_t_allocate_and_get_3d_block_c_dp(tensor, ind, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN)  :: ind
      COMPLEX(kind=real_8), DIMENSION(:,:,:), &
         ALLOCATABLE, INTENT(OUT)                           :: block
      INTEGER, DIMENSION(3)                          :: blk_size

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      CALL allocate_any(block, shape_spec=blk_size)
      CALL dbcsr_t_get_3d_block_c_dp(tensor, ind, blk_size, block)

   END SUBROUTINE
# 700 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   SUBROUTINE dbcsr_t_allocate_and_get_4d_block_c_dp(tensor, ind, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN)  :: ind
      COMPLEX(kind=real_8), DIMENSION(:,:,:,:), &
         ALLOCATABLE, INTENT(OUT)                           :: block
      INTEGER, DIMENSION(4)                          :: blk_size

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      CALL allocate_any(block, shape_spec=blk_size)
      CALL dbcsr_t_get_4d_block_c_dp(tensor, ind, blk_size, block)

   END SUBROUTINE
# 713 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 699 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 700 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   SUBROUTINE dbcsr_t_allocate_and_get_2d_block_c_sp(tensor, ind, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN)  :: ind
      COMPLEX(kind=real_4), DIMENSION(:,:), &
         ALLOCATABLE, INTENT(OUT)                           :: block
      INTEGER, DIMENSION(2)                          :: blk_size

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      CALL allocate_any(block, shape_spec=blk_size)
      CALL dbcsr_t_get_2d_block_c_sp(tensor, ind, blk_size, block)

   END SUBROUTINE
# 700 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   SUBROUTINE dbcsr_t_allocate_and_get_3d_block_c_sp(tensor, ind, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN)  :: ind
      COMPLEX(kind=real_4), DIMENSION(:,:,:), &
         ALLOCATABLE, INTENT(OUT)                           :: block
      INTEGER, DIMENSION(3)                          :: blk_size

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      CALL allocate_any(block, shape_spec=blk_size)
      CALL dbcsr_t_get_3d_block_c_sp(tensor, ind, blk_size, block)

   END SUBROUTINE
# 700 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
   SUBROUTINE dbcsr_t_allocate_and_get_4d_block_c_sp(tensor, ind, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN)  :: ind
      COMPLEX(kind=real_4), DIMENSION(:,:,:,:), &
         ALLOCATABLE, INTENT(OUT)                           :: block
      INTEGER, DIMENSION(4)                          :: blk_size

      CALL dbcsr_t_blk_sizes(tensor, ind, blk_size)
      CALL allocate_any(block, shape_spec=blk_size)
      CALL dbcsr_t_get_4d_block_c_sp(tensor, ind, blk_size, block)

   END SUBROUTINE
# 713 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 714 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"


# 717 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 718 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_get_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_2d_block_r_dp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      REAL(kind=real_8), DIMENSION(sizes(1), sizes(2)), &
         INTENT(OUT)                                        :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d
      REAL(kind=real_8), DIMENSION(:,:), POINTER                    :: block_2d_ptr
      REAL(kind=real_8), DIMENSION(:,:), ALLOCATABLE                :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      LOGICAL                                               :: found
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_get_2d_block_r_dp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)

      NULLIFY (block_2d_ptr)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)
      CALL dbcsr_get_block_p(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d_ptr, found)
      CPASSERT(found)
      ! convert pointer to allocatable
      CALL allocate_any(block_2d, source=block_2d_ptr)

      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL reshape_2d_to_nd_block(map_blk, block_2d, block)

      CALL timestop(handle)
   END SUBROUTINE
# 718 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_get_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_3d_block_r_dp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      REAL(kind=real_8), DIMENSION(sizes(1), sizes(2), sizes(3)), &
         INTENT(OUT)                                        :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d
      REAL(kind=real_8), DIMENSION(:,:), POINTER                    :: block_2d_ptr
      REAL(kind=real_8), DIMENSION(:,:), ALLOCATABLE                :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      LOGICAL                                               :: found
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_get_3d_block_r_dp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)

      NULLIFY (block_2d_ptr)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)
      CALL dbcsr_get_block_p(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d_ptr, found)
      CPASSERT(found)
      ! convert pointer to allocatable
      CALL allocate_any(block_2d, source=block_2d_ptr)

      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL reshape_2d_to_nd_block(map_blk, block_2d, block)

      CALL timestop(handle)
   END SUBROUTINE
# 718 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_get_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_4d_block_r_dp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      REAL(kind=real_8), DIMENSION(sizes(1), sizes(2), sizes(3), sizes(4)), &
         INTENT(OUT)                                        :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d
      REAL(kind=real_8), DIMENSION(:,:), POINTER                    :: block_2d_ptr
      REAL(kind=real_8), DIMENSION(:,:), ALLOCATABLE                :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      LOGICAL                                               :: found
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_get_4d_block_r_dp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)

      NULLIFY (block_2d_ptr)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)
      CALL dbcsr_get_block_p(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d_ptr, found)
      CPASSERT(found)
      ! convert pointer to allocatable
      CALL allocate_any(block_2d, source=block_2d_ptr)

      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL reshape_2d_to_nd_block(map_blk, block_2d, block)

      CALL timestop(handle)
   END SUBROUTINE
# 760 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 717 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 718 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_get_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_2d_block_r_sp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      REAL(kind=real_4), DIMENSION(sizes(1), sizes(2)), &
         INTENT(OUT)                                        :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d
      REAL(kind=real_4), DIMENSION(:,:), POINTER                    :: block_2d_ptr
      REAL(kind=real_4), DIMENSION(:,:), ALLOCATABLE                :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      LOGICAL                                               :: found
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_get_2d_block_r_sp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)

      NULLIFY (block_2d_ptr)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)
      CALL dbcsr_get_block_p(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d_ptr, found)
      CPASSERT(found)
      ! convert pointer to allocatable
      CALL allocate_any(block_2d, source=block_2d_ptr)

      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL reshape_2d_to_nd_block(map_blk, block_2d, block)

      CALL timestop(handle)
   END SUBROUTINE
# 718 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_get_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_3d_block_r_sp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      REAL(kind=real_4), DIMENSION(sizes(1), sizes(2), sizes(3)), &
         INTENT(OUT)                                        :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d
      REAL(kind=real_4), DIMENSION(:,:), POINTER                    :: block_2d_ptr
      REAL(kind=real_4), DIMENSION(:,:), ALLOCATABLE                :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      LOGICAL                                               :: found
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_get_3d_block_r_sp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)

      NULLIFY (block_2d_ptr)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)
      CALL dbcsr_get_block_p(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d_ptr, found)
      CPASSERT(found)
      ! convert pointer to allocatable
      CALL allocate_any(block_2d, source=block_2d_ptr)

      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL reshape_2d_to_nd_block(map_blk, block_2d, block)

      CALL timestop(handle)
   END SUBROUTINE
# 718 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_get_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_4d_block_r_sp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      REAL(kind=real_4), DIMENSION(sizes(1), sizes(2), sizes(3), sizes(4)), &
         INTENT(OUT)                                        :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d
      REAL(kind=real_4), DIMENSION(:,:), POINTER                    :: block_2d_ptr
      REAL(kind=real_4), DIMENSION(:,:), ALLOCATABLE                :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      LOGICAL                                               :: found
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_get_4d_block_r_sp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)

      NULLIFY (block_2d_ptr)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)
      CALL dbcsr_get_block_p(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d_ptr, found)
      CPASSERT(found)
      ! convert pointer to allocatable
      CALL allocate_any(block_2d, source=block_2d_ptr)

      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL reshape_2d_to_nd_block(map_blk, block_2d, block)

      CALL timestop(handle)
   END SUBROUTINE
# 760 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 717 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 718 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_get_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_2d_block_c_dp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      COMPLEX(kind=real_8), DIMENSION(sizes(1), sizes(2)), &
         INTENT(OUT)                                        :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d
      COMPLEX(kind=real_8), DIMENSION(:,:), POINTER                    :: block_2d_ptr
      COMPLEX(kind=real_8), DIMENSION(:,:), ALLOCATABLE                :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      LOGICAL                                               :: found
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_get_2d_block_c_dp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)

      NULLIFY (block_2d_ptr)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)
      CALL dbcsr_get_block_p(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d_ptr, found)
      CPASSERT(found)
      ! convert pointer to allocatable
      CALL allocate_any(block_2d, source=block_2d_ptr)

      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL reshape_2d_to_nd_block(map_blk, block_2d, block)

      CALL timestop(handle)
   END SUBROUTINE
# 718 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_get_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_3d_block_c_dp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      COMPLEX(kind=real_8), DIMENSION(sizes(1), sizes(2), sizes(3)), &
         INTENT(OUT)                                        :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d
      COMPLEX(kind=real_8), DIMENSION(:,:), POINTER                    :: block_2d_ptr
      COMPLEX(kind=real_8), DIMENSION(:,:), ALLOCATABLE                :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      LOGICAL                                               :: found
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_get_3d_block_c_dp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)

      NULLIFY (block_2d_ptr)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)
      CALL dbcsr_get_block_p(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d_ptr, found)
      CPASSERT(found)
      ! convert pointer to allocatable
      CALL allocate_any(block_2d, source=block_2d_ptr)

      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL reshape_2d_to_nd_block(map_blk, block_2d, block)

      CALL timestop(handle)
   END SUBROUTINE
# 718 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_get_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_4d_block_c_dp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      COMPLEX(kind=real_8), DIMENSION(sizes(1), sizes(2), sizes(3), sizes(4)), &
         INTENT(OUT)                                        :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d
      COMPLEX(kind=real_8), DIMENSION(:,:), POINTER                    :: block_2d_ptr
      COMPLEX(kind=real_8), DIMENSION(:,:), ALLOCATABLE                :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      LOGICAL                                               :: found
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_get_4d_block_c_dp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)

      NULLIFY (block_2d_ptr)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)
      CALL dbcsr_get_block_p(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d_ptr, found)
      CPASSERT(found)
      ! convert pointer to allocatable
      CALL allocate_any(block_2d, source=block_2d_ptr)

      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL reshape_2d_to_nd_block(map_blk, block_2d, block)

      CALL timestop(handle)
   END SUBROUTINE
# 760 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 717 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 718 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_get_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_2d_block_c_sp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      COMPLEX(kind=real_4), DIMENSION(sizes(1), sizes(2)), &
         INTENT(OUT)                                        :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d
      COMPLEX(kind=real_4), DIMENSION(:,:), POINTER                    :: block_2d_ptr
      COMPLEX(kind=real_4), DIMENSION(:,:), ALLOCATABLE                :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      LOGICAL                                               :: found
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_get_2d_block_c_sp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)

      NULLIFY (block_2d_ptr)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)
      CALL dbcsr_get_block_p(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d_ptr, found)
      CPASSERT(found)
      ! convert pointer to allocatable
      CALL allocate_any(block_2d, source=block_2d_ptr)

      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL reshape_2d_to_nd_block(map_blk, block_2d, block)

      CALL timestop(handle)
   END SUBROUTINE
# 718 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_get_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_3d_block_c_sp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      COMPLEX(kind=real_4), DIMENSION(sizes(1), sizes(2), sizes(3)), &
         INTENT(OUT)                                        :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d
      COMPLEX(kind=real_4), DIMENSION(:,:), POINTER                    :: block_2d_ptr
      COMPLEX(kind=real_4), DIMENSION(:,:), ALLOCATABLE                :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      LOGICAL                                               :: found
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_get_3d_block_c_sp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)

      NULLIFY (block_2d_ptr)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)
      CALL dbcsr_get_block_p(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d_ptr, found)
      CPASSERT(found)
      ! convert pointer to allocatable
      CALL allocate_any(block_2d, source=block_2d_ptr)

      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL reshape_2d_to_nd_block(map_blk, block_2d, block)

      CALL timestop(handle)
   END SUBROUTINE
# 718 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief Template for dbcsr_t_get_block.
!> \param tensor ...
!> \param ind block index
!> \param sizes block size
!> \param block block to get
! **************************************************************************************************
   SUBROUTINE dbcsr_t_get_4d_block_c_sp(tensor, ind, sizes, block)
      TYPE(dbcsr_t_type), INTENT(INOUT)                     :: tensor
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: ind
      INTEGER, DIMENSION(ndims_tensor(tensor)), INTENT(IN) :: sizes
      COMPLEX(kind=real_4), DIMENSION(sizes(1), sizes(2), sizes(3), sizes(4)), &
         INTENT(OUT)                                        :: block

      INTEGER, ALLOCATABLE, DIMENSION(:)                    :: map1_2d, map2_2d
      INTEGER, DIMENSION(2)                                 :: ind_2d
      COMPLEX(kind=real_4), DIMENSION(:,:), POINTER                    :: block_2d_ptr
      COMPLEX(kind=real_4), DIMENSION(:,:), ALLOCATABLE                :: block_2d
      TYPE(nd_to_2d_mapping)                                :: map_blk
      LOGICAL                                               :: found
      INTEGER                                   :: handle
      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_t_get_4d_block_c_sp', &
         routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)

      NULLIFY (block_2d_ptr)

      ! convert block index
      ind_2d(:) = get_2d_indices(tensor%nd_index_blk, ind)
      CALL dbcsr_get_block_p(tensor%matrix_rep, ind_2d(1), ind_2d(2), block_2d_ptr, found)
      CPASSERT(found)
      ! convert pointer to allocatable
      CALL allocate_any(block_2d, source=block_2d_ptr)

      CALL get_mapping_info(tensor%nd_index_blk, map1_2d=map1_2d, map2_2d=map2_2d)
      CALL create_nd_to_2d_mapping(map_blk, sizes, map1_2d, map2_2d)
      CALL reshape_2d_to_nd_block(map_blk, block_2d, block)

      CALL timestop(handle)
   END SUBROUTINE
# 760 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 761 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

# 763 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 764 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_m2t_2d_r_dp(map, block_2d, block)
      TYPE(nd_to_2d_mapping), INTENT(IN)                        :: map
      REAL(kind=real_8), DIMENSION(:, :), INTENT(IN)                    :: block_2d
      REAL(kind=real_8), DIMENSION(map%dims_nd(1), map%dims_nd(2)), &
         INTENT(OUT)                                            :: block
      INTEGER, DIMENSION(ndims_mapping(map))                    :: map_tmp

      map_tmp(:) = [map%map1_2d, map%map2_2d]
      block(:,:) = RESHAPE(block_2d, SHAPE=SHAPE(block), ORDER=map_tmp)
   END SUBROUTINE
# 764 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_m2t_3d_r_dp(map, block_2d, block)
      TYPE(nd_to_2d_mapping), INTENT(IN)                        :: map
      REAL(kind=real_8), DIMENSION(:, :), INTENT(IN)                    :: block_2d
      REAL(kind=real_8), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3)), &
         INTENT(OUT)                                            :: block
      INTEGER, DIMENSION(ndims_mapping(map))                    :: map_tmp

      map_tmp(:) = [map%map1_2d, map%map2_2d]
      block(:,:,:) = RESHAPE(block_2d, SHAPE=SHAPE(block), ORDER=map_tmp)
   END SUBROUTINE
# 764 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_m2t_4d_r_dp(map, block_2d, block)
      TYPE(nd_to_2d_mapping), INTENT(IN)                        :: map
      REAL(kind=real_8), DIMENSION(:, :), INTENT(IN)                    :: block_2d
      REAL(kind=real_8), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3), map%dims_nd(4)), &
         INTENT(OUT)                                            :: block
      INTEGER, DIMENSION(ndims_mapping(map))                    :: map_tmp

      map_tmp(:) = [map%map1_2d, map%map2_2d]
      block(:,:,:,:) = RESHAPE(block_2d, SHAPE=SHAPE(block), ORDER=map_tmp)
   END SUBROUTINE
# 782 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 763 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 764 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_m2t_2d_r_sp(map, block_2d, block)
      TYPE(nd_to_2d_mapping), INTENT(IN)                        :: map
      REAL(kind=real_4), DIMENSION(:, :), INTENT(IN)                    :: block_2d
      REAL(kind=real_4), DIMENSION(map%dims_nd(1), map%dims_nd(2)), &
         INTENT(OUT)                                            :: block
      INTEGER, DIMENSION(ndims_mapping(map))                    :: map_tmp

      map_tmp(:) = [map%map1_2d, map%map2_2d]
      block(:,:) = RESHAPE(block_2d, SHAPE=SHAPE(block), ORDER=map_tmp)
   END SUBROUTINE
# 764 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_m2t_3d_r_sp(map, block_2d, block)
      TYPE(nd_to_2d_mapping), INTENT(IN)                        :: map
      REAL(kind=real_4), DIMENSION(:, :), INTENT(IN)                    :: block_2d
      REAL(kind=real_4), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3)), &
         INTENT(OUT)                                            :: block
      INTEGER, DIMENSION(ndims_mapping(map))                    :: map_tmp

      map_tmp(:) = [map%map1_2d, map%map2_2d]
      block(:,:,:) = RESHAPE(block_2d, SHAPE=SHAPE(block), ORDER=map_tmp)
   END SUBROUTINE
# 764 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_m2t_4d_r_sp(map, block_2d, block)
      TYPE(nd_to_2d_mapping), INTENT(IN)                        :: map
      REAL(kind=real_4), DIMENSION(:, :), INTENT(IN)                    :: block_2d
      REAL(kind=real_4), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3), map%dims_nd(4)), &
         INTENT(OUT)                                            :: block
      INTEGER, DIMENSION(ndims_mapping(map))                    :: map_tmp

      map_tmp(:) = [map%map1_2d, map%map2_2d]
      block(:,:,:,:) = RESHAPE(block_2d, SHAPE=SHAPE(block), ORDER=map_tmp)
   END SUBROUTINE
# 782 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 763 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 764 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_m2t_2d_c_dp(map, block_2d, block)
      TYPE(nd_to_2d_mapping), INTENT(IN)                        :: map
      COMPLEX(kind=real_8), DIMENSION(:, :), INTENT(IN)                    :: block_2d
      COMPLEX(kind=real_8), DIMENSION(map%dims_nd(1), map%dims_nd(2)), &
         INTENT(OUT)                                            :: block
      INTEGER, DIMENSION(ndims_mapping(map))                    :: map_tmp

      map_tmp(:) = [map%map1_2d, map%map2_2d]
      block(:,:) = RESHAPE(block_2d, SHAPE=SHAPE(block), ORDER=map_tmp)
   END SUBROUTINE
# 764 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_m2t_3d_c_dp(map, block_2d, block)
      TYPE(nd_to_2d_mapping), INTENT(IN)                        :: map
      COMPLEX(kind=real_8), DIMENSION(:, :), INTENT(IN)                    :: block_2d
      COMPLEX(kind=real_8), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3)), &
         INTENT(OUT)                                            :: block
      INTEGER, DIMENSION(ndims_mapping(map))                    :: map_tmp

      map_tmp(:) = [map%map1_2d, map%map2_2d]
      block(:,:,:) = RESHAPE(block_2d, SHAPE=SHAPE(block), ORDER=map_tmp)
   END SUBROUTINE
# 764 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_m2t_4d_c_dp(map, block_2d, block)
      TYPE(nd_to_2d_mapping), INTENT(IN)                        :: map
      COMPLEX(kind=real_8), DIMENSION(:, :), INTENT(IN)                    :: block_2d
      COMPLEX(kind=real_8), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3), map%dims_nd(4)), &
         INTENT(OUT)                                            :: block
      INTEGER, DIMENSION(ndims_mapping(map))                    :: map_tmp

      map_tmp(:) = [map%map1_2d, map%map2_2d]
      block(:,:,:,:) = RESHAPE(block_2d, SHAPE=SHAPE(block), ORDER=map_tmp)
   END SUBROUTINE
# 782 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 763 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 764 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_m2t_2d_c_sp(map, block_2d, block)
      TYPE(nd_to_2d_mapping), INTENT(IN)                        :: map
      COMPLEX(kind=real_4), DIMENSION(:, :), INTENT(IN)                    :: block_2d
      COMPLEX(kind=real_4), DIMENSION(map%dims_nd(1), map%dims_nd(2)), &
         INTENT(OUT)                                            :: block
      INTEGER, DIMENSION(ndims_mapping(map))                    :: map_tmp

      map_tmp(:) = [map%map1_2d, map%map2_2d]
      block(:,:) = RESHAPE(block_2d, SHAPE=SHAPE(block), ORDER=map_tmp)
   END SUBROUTINE
# 764 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_m2t_3d_c_sp(map, block_2d, block)
      TYPE(nd_to_2d_mapping), INTENT(IN)                        :: map
      COMPLEX(kind=real_4), DIMENSION(:, :), INTENT(IN)                    :: block_2d
      COMPLEX(kind=real_4), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3)), &
         INTENT(OUT)                                            :: block
      INTEGER, DIMENSION(ndims_mapping(map))                    :: map_tmp

      map_tmp(:) = [map%map1_2d, map%map2_2d]
      block(:,:,:) = RESHAPE(block_2d, SHAPE=SHAPE(block), ORDER=map_tmp)
   END SUBROUTINE
# 764 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_m2t_4d_c_sp(map, block_2d, block)
      TYPE(nd_to_2d_mapping), INTENT(IN)                        :: map
      COMPLEX(kind=real_4), DIMENSION(:, :), INTENT(IN)                    :: block_2d
      COMPLEX(kind=real_4), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3), map%dims_nd(4)), &
         INTENT(OUT)                                            :: block
      INTEGER, DIMENSION(ndims_mapping(map))                    :: map_tmp

      map_tmp(:) = [map%map1_2d, map%map2_2d]
      block(:,:,:,:) = RESHAPE(block_2d, SHAPE=SHAPE(block), ORDER=map_tmp)
   END SUBROUTINE
# 782 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 783 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

# 785 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 786 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_t2m_2d_r_dp(map, block, block_2d)
      TYPE(nd_to_2d_mapping), INTENT(IN)                                 :: map
      REAL(kind=real_8), DIMENSION(map%dims_nd(1), map%dims_nd(2)), INTENT(IN) :: block
      REAL(kind=real_8), DIMENSION(map%dims_2d(1), map%dims_2d(2)), INTENT(OUT)  :: block_2d

      INTEGER, DIMENSION(ndims_mapping(map))                             :: shape_in, shape_reordered
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:)           :: block_tmp

      shape_in = SHAPE(block)
      shape_reordered(map%map_nd) = shape_in
      CALL allocate_any(block_tmp, source=block, order=map%map_nd)
      block_2d(:, :) = RESHAPE(block_tmp, SHAPE=SHAPE(block_2d))
   END SUBROUTINE
# 786 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_t2m_3d_r_dp(map, block, block_2d)
      TYPE(nd_to_2d_mapping), INTENT(IN)                                 :: map
      REAL(kind=real_8), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3)), INTENT(IN) :: block
      REAL(kind=real_8), DIMENSION(map%dims_2d(1), map%dims_2d(2)), INTENT(OUT)  :: block_2d

      INTEGER, DIMENSION(ndims_mapping(map))                             :: shape_in, shape_reordered
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:)           :: block_tmp

      shape_in = SHAPE(block)
      shape_reordered(map%map_nd) = shape_in
      CALL allocate_any(block_tmp, source=block, order=map%map_nd)
      block_2d(:, :) = RESHAPE(block_tmp, SHAPE=SHAPE(block_2d))
   END SUBROUTINE
# 786 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_t2m_4d_r_dp(map, block, block_2d)
      TYPE(nd_to_2d_mapping), INTENT(IN)                                 :: map
      REAL(kind=real_8), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3), map%dims_nd(4)), INTENT(IN) :: block
      REAL(kind=real_8), DIMENSION(map%dims_2d(1), map%dims_2d(2)), INTENT(OUT)  :: block_2d

      INTEGER, DIMENSION(ndims_mapping(map))                             :: shape_in, shape_reordered
      REAL(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:,:)           :: block_tmp

      shape_in = SHAPE(block)
      shape_reordered(map%map_nd) = shape_in
      CALL allocate_any(block_tmp, source=block, order=map%map_nd)
      block_2d(:, :) = RESHAPE(block_tmp, SHAPE=SHAPE(block_2d))
   END SUBROUTINE
# 806 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 785 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 786 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_t2m_2d_r_sp(map, block, block_2d)
      TYPE(nd_to_2d_mapping), INTENT(IN)                                 :: map
      REAL(kind=real_4), DIMENSION(map%dims_nd(1), map%dims_nd(2)), INTENT(IN) :: block
      REAL(kind=real_4), DIMENSION(map%dims_2d(1), map%dims_2d(2)), INTENT(OUT)  :: block_2d

      INTEGER, DIMENSION(ndims_mapping(map))                             :: shape_in, shape_reordered
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:)           :: block_tmp

      shape_in = SHAPE(block)
      shape_reordered(map%map_nd) = shape_in
      CALL allocate_any(block_tmp, source=block, order=map%map_nd)
      block_2d(:, :) = RESHAPE(block_tmp, SHAPE=SHAPE(block_2d))
   END SUBROUTINE
# 786 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_t2m_3d_r_sp(map, block, block_2d)
      TYPE(nd_to_2d_mapping), INTENT(IN)                                 :: map
      REAL(kind=real_4), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3)), INTENT(IN) :: block
      REAL(kind=real_4), DIMENSION(map%dims_2d(1), map%dims_2d(2)), INTENT(OUT)  :: block_2d

      INTEGER, DIMENSION(ndims_mapping(map))                             :: shape_in, shape_reordered
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:)           :: block_tmp

      shape_in = SHAPE(block)
      shape_reordered(map%map_nd) = shape_in
      CALL allocate_any(block_tmp, source=block, order=map%map_nd)
      block_2d(:, :) = RESHAPE(block_tmp, SHAPE=SHAPE(block_2d))
   END SUBROUTINE
# 786 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_t2m_4d_r_sp(map, block, block_2d)
      TYPE(nd_to_2d_mapping), INTENT(IN)                                 :: map
      REAL(kind=real_4), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3), map%dims_nd(4)), INTENT(IN) :: block
      REAL(kind=real_4), DIMENSION(map%dims_2d(1), map%dims_2d(2)), INTENT(OUT)  :: block_2d

      INTEGER, DIMENSION(ndims_mapping(map))                             :: shape_in, shape_reordered
      REAL(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:,:)           :: block_tmp

      shape_in = SHAPE(block)
      shape_reordered(map%map_nd) = shape_in
      CALL allocate_any(block_tmp, source=block, order=map%map_nd)
      block_2d(:, :) = RESHAPE(block_tmp, SHAPE=SHAPE(block_2d))
   END SUBROUTINE
# 806 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 785 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 786 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_t2m_2d_c_dp(map, block, block_2d)
      TYPE(nd_to_2d_mapping), INTENT(IN)                                 :: map
      COMPLEX(kind=real_8), DIMENSION(map%dims_nd(1), map%dims_nd(2)), INTENT(IN) :: block
      COMPLEX(kind=real_8), DIMENSION(map%dims_2d(1), map%dims_2d(2)), INTENT(OUT)  :: block_2d

      INTEGER, DIMENSION(ndims_mapping(map))                             :: shape_in, shape_reordered
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:)           :: block_tmp

      shape_in = SHAPE(block)
      shape_reordered(map%map_nd) = shape_in
      CALL allocate_any(block_tmp, source=block, order=map%map_nd)
      block_2d(:, :) = RESHAPE(block_tmp, SHAPE=SHAPE(block_2d))
   END SUBROUTINE
# 786 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_t2m_3d_c_dp(map, block, block_2d)
      TYPE(nd_to_2d_mapping), INTENT(IN)                                 :: map
      COMPLEX(kind=real_8), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3)), INTENT(IN) :: block
      COMPLEX(kind=real_8), DIMENSION(map%dims_2d(1), map%dims_2d(2)), INTENT(OUT)  :: block_2d

      INTEGER, DIMENSION(ndims_mapping(map))                             :: shape_in, shape_reordered
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:)           :: block_tmp

      shape_in = SHAPE(block)
      shape_reordered(map%map_nd) = shape_in
      CALL allocate_any(block_tmp, source=block, order=map%map_nd)
      block_2d(:, :) = RESHAPE(block_tmp, SHAPE=SHAPE(block_2d))
   END SUBROUTINE
# 786 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_t2m_4d_c_dp(map, block, block_2d)
      TYPE(nd_to_2d_mapping), INTENT(IN)                                 :: map
      COMPLEX(kind=real_8), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3), map%dims_nd(4)), INTENT(IN) :: block
      COMPLEX(kind=real_8), DIMENSION(map%dims_2d(1), map%dims_2d(2)), INTENT(OUT)  :: block_2d

      INTEGER, DIMENSION(ndims_mapping(map))                             :: shape_in, shape_reordered
      COMPLEX(kind=real_8), ALLOCATABLE, DIMENSION(:,:,:,:)           :: block_tmp

      shape_in = SHAPE(block)
      shape_reordered(map%map_nd) = shape_in
      CALL allocate_any(block_tmp, source=block, order=map%map_nd)
      block_2d(:, :) = RESHAPE(block_tmp, SHAPE=SHAPE(block_2d))
   END SUBROUTINE
# 806 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 785 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 786 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_t2m_2d_c_sp(map, block, block_2d)
      TYPE(nd_to_2d_mapping), INTENT(IN)                                 :: map
      COMPLEX(kind=real_4), DIMENSION(map%dims_nd(1), map%dims_nd(2)), INTENT(IN) :: block
      COMPLEX(kind=real_4), DIMENSION(map%dims_2d(1), map%dims_2d(2)), INTENT(OUT)  :: block_2d

      INTEGER, DIMENSION(ndims_mapping(map))                             :: shape_in, shape_reordered
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:)           :: block_tmp

      shape_in = SHAPE(block)
      shape_reordered(map%map_nd) = shape_in
      CALL allocate_any(block_tmp, source=block, order=map%map_nd)
      block_2d(:, :) = RESHAPE(block_tmp, SHAPE=SHAPE(block_2d))
   END SUBROUTINE
# 786 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_t2m_3d_c_sp(map, block, block_2d)
      TYPE(nd_to_2d_mapping), INTENT(IN)                                 :: map
      COMPLEX(kind=real_4), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3)), INTENT(IN) :: block
      COMPLEX(kind=real_4), DIMENSION(map%dims_2d(1), map%dims_2d(2)), INTENT(OUT)  :: block_2d

      INTEGER, DIMENSION(ndims_mapping(map))                             :: shape_in, shape_reordered
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:)           :: block_tmp

      shape_in = SHAPE(block)
      shape_reordered(map%map_nd) = shape_in
      CALL allocate_any(block_tmp, source=block, order=map%map_nd)
      block_2d(:, :) = RESHAPE(block_tmp, SHAPE=SHAPE(block_2d))
   END SUBROUTINE
# 786 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
! **************************************************************************************************
!> \brief ...
!> \param map ...
!> \param block_2d ...
!> \param block ...
! **************************************************************************************************
   SUBROUTINE reshape_block_t2m_4d_c_sp(map, block, block_2d)
      TYPE(nd_to_2d_mapping), INTENT(IN)                                 :: map
      COMPLEX(kind=real_4), DIMENSION(map%dims_nd(1), map%dims_nd(2), map%dims_nd(3), map%dims_nd(4)), INTENT(IN) :: block
      COMPLEX(kind=real_4), DIMENSION(map%dims_2d(1), map%dims_2d(2)), INTENT(OUT)  :: block_2d

      INTEGER, DIMENSION(ndims_mapping(map))                             :: shape_in, shape_reordered
      COMPLEX(kind=real_4), ALLOCATABLE, DIMENSION(:,:,:,:)           :: block_tmp

      shape_in = SHAPE(block)
      shape_reordered(map%map_nd) = shape_in
      CALL allocate_any(block_tmp, source=block, order=map%map_nd)
      block_2d(:, :) = RESHAPE(block_tmp, SHAPE=SHAPE(block_2d))
   END SUBROUTINE
# 806 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"
# 807 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr_tensor/dbcsr_tensor_block.F"

END MODULE
