# 1 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2018  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief represent a simple array based list of the given type
!> \par History
!>      08.2003 created [fawzi]
!>      01.2017 ported to Fypp [Ole Schuett]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE molecule_kind_list_types
    use molecule_kind_types, only: molecule_kind_type, deallocate_molecule_kind_set

#include "../base/base_uses.f90"

   IMPLICIT NONE
   PRIVATE

# 1 "/builddir/build/BUILD/cp2k-6.1/src/subsys/subsys_list.fypp" 1
# 5 "/builddir/build/BUILD/cp2k-6.1/src/subsys/subsys_list.fypp"
# 152 "/builddir/build/BUILD/cp2k-6.1/src/subsys/subsys_list.fypp"
# 22 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F" 2


# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  LOGICAL, PRIVATE, PARAMETER :: debug_this_module=.TRUE.
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  CHARACTER(len=*), PRIVATE, PARAMETER :: moduleN='molecule_kind_list_types'
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  INTEGER, PRIVATE, SAVE :: last_molecule_kind_list_id=0
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  !API
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  PUBLIC :: molecule_kind_list_type,   molecule_kind_list_p_type,&
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
            molecule_kind_list_create, molecule_kind_list_retain,&
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
            molecule_kind_list_release
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
! **************************************************************************************************
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \brief represent a list of objects
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \param id_nr identification number of this list
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \param ref_count reference count (see doc/ReferenceCounting.html)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \param n_el the number of elements in the list
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \param owns_list if the els are owned by this structure, and
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!>        should be deallocated by it
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \param list the array of object, might be oversized,
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!>        only the fist n_el have some meaning
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \par History
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!>      08.2003 created [fawzi]
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \author Fawzi Mohamed
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
! **************************************************************************************************
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  TYPE molecule_kind_list_type
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
     INTEGER :: id_nr, ref_count, n_els
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
     LOGICAL :: owns_els
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
     TYPE(molecule_kind_type), DIMENSION(:), POINTER :: els
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  END TYPE molecule_kind_list_type
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
! **************************************************************************************************
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \brief represents a pointer to a list
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \param list the pointer to the list
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \par History
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!>      08.2003 created [fawzi]
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \author Fawzi Mohamed
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
! **************************************************************************************************
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  TYPE molecule_kind_list_p_type
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
     TYPE(molecule_kind_list_type), POINTER :: list
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  END TYPE molecule_kind_list_p_type
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
CONTAINS
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
! **************************************************************************************************
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \brief creates a list
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \param list the list to allocate and initialize
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \param els_ptr the elements to store in the list (the array is only,
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!>        referenced, not copied!)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \param owns_els if the list takes the ownership of els_ptr and
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!>        will deallocate it (defaults to true)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \param n_els number of elements in the list (at least one els_ptr or
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!>        n_els should be given)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \par History
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!>      08.2003 created [fawzi]
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \author Fawzi Mohamed
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
! **************************************************************************************************
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
SUBROUTINE molecule_kind_list_create(list, els_ptr, &
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
     owns_els, n_els)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  TYPE(molecule_kind_list_type), POINTER, OPTIONAL :: list
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  TYPE(molecule_kind_type), DIMENSION(:), POINTER, OPTIONAL :: els_ptr
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  LOGICAL, INTENT(in), OPTIONAL :: owns_els
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  INTEGER, INTENT(in), OPTIONAL :: n_els
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  CHARACTER(len=*), PARAMETER :: routineN='molecule_kind_list_create',&
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
        routineP=moduleN//':'//routineN
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  CPASSERT(PRESENT(els_ptr).OR.PRESENT(n_els))
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  ALLOCATE(list)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  last_molecule_kind_list_id=last_molecule_kind_list_id+1
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  list%id_nr=last_molecule_kind_list_id
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  list%ref_count=1
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  list%owns_els=.TRUE.
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  list%n_els=0
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  IF (PRESENT(owns_els)) list%owns_els=owns_els
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  NULLIFY(list%els)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  IF (PRESENT(els_ptr)) THEN
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
     list%els => els_ptr
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
     IF (ASSOCIATED(els_ptr)) THEN
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
        list%n_els=SIZE(els_ptr)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
     END IF
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  END IF
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  IF (PRESENT(n_els)) list%n_els=n_els
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  IF (.NOT.ASSOCIATED(list%els)) THEN
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
     ALLOCATE(list%els(list%n_els))
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
     CPASSERT(list%owns_els)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  END IF
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
END SUBROUTINE molecule_kind_list_create
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
! **************************************************************************************************
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \brief retains a list (see doc/ReferenceCounting.html)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \param list the list to retain
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \par History
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!>      08.2003 created [fawzi]
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \author Fawzi Mohamed
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
! **************************************************************************************************
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
SUBROUTINE molecule_kind_list_retain(list)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  TYPE(molecule_kind_list_type), POINTER :: list
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  CHARACTER(len=*), PARAMETER :: routineN='molecule_kind_list_retain',&
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
        routineP=moduleN//':'//routineN
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  CPASSERT(ASSOCIATED(list))
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  CPASSERT(list%ref_count>0)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  list%ref_count=list%ref_count+1
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
END SUBROUTINE molecule_kind_list_retain
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
! **************************************************************************************************
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \brief releases a  list (see doc/ReferenceCounting.html)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \param list the list to release
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \par History
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!>      08.2003 created [fawzi]
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
!> \author Fawzi Mohamed
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
! **************************************************************************************************
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
SUBROUTINE molecule_kind_list_release(list)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  TYPE(molecule_kind_list_type), POINTER :: list
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  CHARACTER(len=*), PARAMETER :: routineN='molecule_kind_list_release',&
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
        routineP=moduleN//':'//routineN
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"

# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  IF (ASSOCIATED(list)) THEN
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
     CPASSERT(list%ref_count>0)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
     list%ref_count=list%ref_count-1
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
     IF (list%ref_count==0) THEN
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
        IF (list%owns_els) THEN
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
           IF (ASSOCIATED(list%els)) THEN
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
              CALL deallocate_molecule_kind_set(list%els)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
           END IF
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
        END IF
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
        NULLIFY(list%els)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
        DEALLOCATE(list)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
     END IF
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  END IF
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
  NULLIFY(list)
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"
END SUBROUTINE molecule_kind_list_release
# 23 "/builddir/build/BUILD/cp2k-6.1/src/subsys/molecule_kind_list_types.F"


END MODULE molecule_kind_list_types
